from eole.utils.logging import logger
from eole.transforms import register_transform
from eole.constants import TransformType
from .transform import Transform, ObservableStats, TransformConfig
from pydantic import Field


class FilterTooShortStats(ObservableStats):
    """Runing statistics for FilterTooShortTransform."""

    __slots__ = ["filtered"]

    def __init__(self):
        self.filtered = 1

    def update(self, other: "FilterTooShortStats"):
        self.filtered += other.filtered


class FilterTooShortConfig(TransformConfig):
    src_seq_length: int | None = Field(default=48, description="Minimum source sequence length.")
    tgt_seq_length: int | None = Field(default=48, description="Minimum target sequence length.")


class FilterTooLongStats(ObservableStats):
    """Runing statistics for FilterTooLongTransform."""

    __slots__ = ["filtered"]

    def __init__(self):
        self.filtered = 1

    def update(self, other: "FilterTooLongStats"):
        self.filtered += other.filtered


class FilterTooLongConfig(TransformConfig):
    src_seq_length: int | None = Field(default=192, description="Maximum source sequence length.")
    tgt_seq_length: int | None = Field(default=192, description="Maximum target sequence length.")


class PrefixConfig(TransformConfig):
    src_prefix: str | None = Field(default="", description="String to prepend to all source examples.")
    tgt_prefix: str | None = Field(default="", description="String to prepend to all target examples.")


class SuffixConfig(TransformConfig):
    src_suffix: str | None = Field(default="", description="String to append to all source examples.")
    tgt_suffix: str | None = Field(default="", description="String to append to all target examples.")


@register_transform(name="filtertooshort")
class FilterTooShortTransform(Transform):
    """Filter out sentence that are too short."""

    config_model = FilterTooShortConfig
    type = TransformType.Train

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        self.src_seq_length = self.config.src_seq_length
        self.tgt_seq_length = self.config.tgt_seq_length

    def apply(self, example, is_train=False, stats=None, **kwargs):
        """Return None if too short else return as is."""
        if len(example["src"]) < self.src_seq_length or (
            example["tgt"] is not None and len(example["tgt"]) < self.tgt_seq_length - 2
        ):
            if stats is not None:
                stats.update(FilterTooLongStats())
            return None
        else:
            return example

    def _repr_args(self):
        """Return str represent key arguments for class."""
        return "{}={}, {}={}".format("src_seq_length", self.src_seq_length, "tgt_seq_length", self.tgt_seq_length)


@register_transform(name="filtertoolong")
class FilterTooLongTransform(Transform):
    """Filter out sentence that are too long."""

    config_model = FilterTooLongConfig
    type = TransformType.Train

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        self.src_seq_length = self.config.src_seq_length
        self.tgt_seq_length = self.config.tgt_seq_length

    def apply(self, example, is_train=False, stats=None, **kwargs):
        """Return None if too long else return as is."""
        if len(example["src"]) > self.src_seq_length or (
            example["tgt"] is not None and len(example["tgt"]) > self.tgt_seq_length - 2
        ):
            if stats is not None:
                stats.update(FilterTooLongStats())
            return None
        else:
            return example

    def _repr_args(self):
        """Return str represent key arguments for class."""
        return "{}={}, {}={}".format("src_seq_length", self.src_seq_length, "tgt_seq_length", self.tgt_seq_length)


@register_transform(name="prefix")
class PrefixTransform(Transform):
    """Add Prefix to src (& tgt) sentence."""

    config_model = PrefixConfig

    def __init__(self, config):
        super().__init__(config)
        self.prefix_dict = {}

    @staticmethod
    def _get_prefix(corpus):
        """Get prefix string of a `corpus`."""
        if "prefix" in corpus.transforms:
            src_prefix = corpus.src_prefix
            tgt_prefix = corpus.tgt_prefix
            prefix = {"src": src_prefix, "tgt": tgt_prefix}
        else:
            prefix = None
        return prefix

    @classmethod
    def get_prefix_dict(cls, config):
        """Get all needed prefix correspond to corpus in `opts`."""
        prefix_dict = {}
        # prefix src/tgt for each dataset
        if hasattr(config, "data"):
            for c_name, corpus in config.data.items():
                prefix = cls._get_prefix(corpus)
                if prefix is not None:
                    logger.debug(f"Get prefix for {c_name}: {prefix}")
                    prefix_dict[c_name] = prefix
        # prefix as general option for inference
        if getattr(config.transforms_configs.prefix, "src_prefix", None) is not None:
            if "infer" not in prefix_dict.keys():
                prefix_dict["infer"] = {}
            prefix_dict["infer"]["src"] = config.transforms_configs.prefix.src_prefix
            logger.debug(f"Get prefix for src infer: {config.transforms_configs.prefix.src_prefix}")
        if getattr(config.transforms_configs.prefix, "tgt_prefix", None) is not None:
            if "infer" not in prefix_dict.keys():
                prefix_dict["infer"] = {}
            prefix_dict["infer"]["tgt"] = config.transforms_configs.prefix.tgt_prefix
            logger.debug(f"Get prefix for tgt infer: {config.transforms_configs.prefix.tgt_prefix}")
        return prefix_dict

    @classmethod
    def get_specials(cls, config):
        """Get special vocabs added by prefix transform."""
        prefix_dict = cls.get_prefix_dict(config)
        src_specials, tgt_specials = set(), set()
        for _, prefix in prefix_dict.items():
            src_specials.update(prefix["src"].split(" "))
            tgt_specials.update(prefix["tgt"].split(" "))
        # we might want to handle the "no prefix" case in a better way
        if "" in src_specials:
            src_specials.remove("")
        if "" in tgt_specials:
            tgt_specials.remove("")
        return (src_specials, tgt_specials)

    def warm_up(self, vocabs=None):
        """Warm up to get prefix dictionary."""
        super().warm_up(None)
        self.prefix_dict = self.get_prefix_dict(self.full_config)

    def _prepend(self, example, prefix):
        """Prepend `prefix` to `tokens`."""
        for side, side_prefix in prefix.items():
            if example.get(side) is not None and len(side_prefix) > 0:
                example[side] = side_prefix.split(" ") + example[side]
            elif len(side_prefix) > 0:
                example[side] = side_prefix.split(" ")
        return example

    def apply(self, example, is_train=False, stats=None, **kwargs):
        """Apply prefix prepend to example.

        Should provide `corpus_name` to get correspond prefix.
        """
        corpus_name = kwargs.get("corpus_name", None)
        if corpus_name is None:
            raise ValueError("corpus_name is required.")
        corpus_prefix = self.prefix_dict.get(corpus_name, None)
        if corpus_prefix is None:
            raise ValueError(f"prefix for {corpus_name} does not exist.")
        return self._prepend(example, corpus_prefix)

    def apply_reverse(self, predicted):
        def _removeprefix(s, prefix):
            if len(prefix) > 0 and s.startswith(prefix):
                return s[len(prefix) + 1 :]
            else:
                return s

        corpus_prefix = self.prefix_dict.get("infer", None)
        return _removeprefix(predicted, corpus_prefix["tgt"])

    def _repr_args(self):
        """Return str represent key arguments for class."""
        return "{}={}".format("prefix_dict", self.prefix_dict)


@register_transform(name="suffix")
class SuffixTransform(Transform):
    """Add Suffix to src (& tgt) sentence."""

    config_model = SuffixConfig

    def __init__(self, config):
        super().__init__(config)

    @staticmethod
    def _get_suffix(corpus):
        """Get suffix string of a `corpus`."""
        if "suffix" in corpus.transforms:
            src_suffix = corpus.src_suffix
            tgt_suffix = corpus.tgt_suffix
            suffix = {"src": src_suffix, "tgt": tgt_suffix}
        else:
            suffix = None
        return suffix

    @classmethod
    def get_suffix_dict(cls, config):
        """Get all needed suffix correspond to corpus in `opts`."""
        suffix_dict = {}
        # suffix src/tgt for each dataset
        if hasattr(config, "data"):
            for c_name, corpus in config.data.items():
                suffix = cls._get_suffix(corpus)
                if suffix is not None:
                    logger.debug(f"Get suffix for {c_name}: {suffix}")
                    suffix_dict[c_name] = suffix
        # suffix as general option for inference
        if hasattr(config.transforms_configs.prefix, "src_suffix"):
            if "infer" not in suffix_dict.keys():
                suffix_dict["infer"] = {}
            suffix_dict["infer"]["src"] = config.transforms_configs.prefix.src_suffix
            logger.debug(f"Get suffix for src infer: {config.transforms_configs.prefix.src_suffix}")
        if hasattr(config.transforms_configs.prefix, "tgt_suffix"):
            if "infer" not in suffix_dict.keys():
                suffix_dict["infer"] = {}
            suffix_dict["infer"]["tgt"] = config.transforms_configs.prefix.tgt_suffix
            logger.debug(f"Get suffix for tgt infer: {config.transforms_configs.prefix.tgt_suffix}")

        return suffix_dict

    @classmethod
    def get_specials(cls, config):
        """Get special vocabs added by suffix transform."""
        suffix_dict = cls.get_suffix_dict(config)
        src_specials, tgt_specials = set(), set()
        for _, suffix in suffix_dict.items():
            src_specials.update(suffix["src"].split(" "))
            tgt_specials.update(suffix["tgt"].split(" "))
        src_specials.remove("")
        tgt_specials.remove("")
        return (src_specials, tgt_specials)

    def warm_up(self, vocabs=None):
        """Warm up to get suffix dictionary."""
        super().warm_up(None)
        self.suffix_dict = self.get_suffix_dict(self.config)

    def _append(self, example, suffix):
        """Prepend `suffix` to `tokens`."""
        for side, side_suffix in suffix.items():
            if example.get(side) is not None and len(side_suffix) > 0:
                example[side] = example[side] + side_suffix.split(" ")
            elif len(side_suffix) > 0:
                example[side] = side_suffix.split(" ")
        return example

    def apply(self, example, is_train=False, stats=None, **kwargs):
        """Apply suffix append to example.

        Should provide `corpus_name` to get correspond suffix.
        """
        corpus_name = kwargs.get("corpus_name", None)
        if corpus_name is None:
            raise ValueError("corpus_name is required.")
        corpus_suffix = self.suffix_dict.get(corpus_name, None)
        if corpus_suffix is None:
            raise ValueError(f"suffix for {corpus_name} does not exist.")
        return self._append(example, corpus_suffix)

    def _repr_args(self):
        """Return str represent key arguments for class."""
        return "{}={}".format("suffix_dict", self.suffix_dict)



"""Module defining decoders."""

from eole.decoders.rnn_decoder import InputFeedRNNDecoder, StdRNNDecoder
from eole.decoders.transformer import TransformerDecoder
from eole.decoders.cnn_decoder import CNNDecoder


str2dec = {
    "rnn": StdRNNDecoder,
    "ifrnn": InputFeedRNNDecoder,
    "cnn": CNNDecoder,
    "transformer": TransformerDecoder,
}



"""
This is the loadable seq2seq trainer library that is
in charge of training details, loss compute, and statistics.
See train.py for a use case of this library.

Note: To make this a general library, we implement *only*
      mechanism things here(i.e. what to do), and leave the strategy
      things to users(i.e. how to do it). Also see train.py(one of the
      users of this library) for the strategy things we do.
"""

import time
import sys
import torch
import traceback
import eole.utils
from eole.utils.distributed import all_gather_list, all_reduce_and_rescale_tensors
from eole.utils.loss import LossCompute
from eole.utils.logging import logger
from eole.utils.misc import clear_gpu_cache, get_autocast
from eole.utils.scoring_utils import ScoringPreparator
from eole.scorers import get_scorers_cls, build_scorers


def build_trainer(config, device_id, model, vocabs, optim, model_saver=None):
    """
    Simplify `Trainer` creation based on user `opt`s*

    Args:
        opt (:obj:`Namespace`): user options (usually from argument parsing)
        model (:obj:`eole.models.BaseModel`): the model to train
        fields (dict): dict of fields
        optim (:obj:`eole.utils.Optimizer`): optimizer used during training
        data_type (str): string describing the type of data
            e.g. "text"
        model_saver(:obj:`eole.models.ModelSaverBase`): the utility object
            used to save the model
    """

    train_loss = LossCompute.from_config(config, model, vocabs)
    valid_loss = LossCompute.from_config(config, model, vocabs, train=False)
    estim_loss_lambda = config.training.estim_loss_lambda
    estim_loss_lambda_steps = config.training.estim_loss_lambda_steps

    scoring_preparator = ScoringPreparator(vocabs, config)
    validset_transforms = getattr(config.data.get("valid", None), "transforms", None)
    if validset_transforms:
        scoring_preparator.warm_up(validset_transforms)
    scorers_cls = get_scorers_cls(config.valid_metrics)
    valid_scorers = build_scorers(config, scorers_cls)
    running_config = config.training
    norm_method = running_config.normalization
    accum_count = running_config.accum_count
    accum_steps = running_config.accum_steps
    n_gpu = running_config.world_size
    parallel_mode = running_config.parallel_mode
    average_decay = running_config.average_decay
    average_every = running_config.average_every
    dropout = running_config.dropout
    attention_dropout = running_config.attention_dropout
    dropout_steps = running_config.dropout_steps
    zero_out_prompt_loss = running_config.zero_out_prompt_loss
    if device_id >= 0:
        gpu_rank = running_config.gpu_ranks[device_id]
    else:
        gpu_rank = -1
        n_gpu = 0

    earlystopper = (
        eole.utils.EarlyStopping(
            running_config.early_stopping,
            scorers=eole.utils.scorers_from_config(running_config),
        )
        if running_config.early_stopping > 0
        else None
    )

    report_manager = eole.utils.build_report_manager(config, gpu_rank)

    trainer = Trainer(
        model,
        train_loss,
        valid_loss,
        scoring_preparator,
        valid_scorers,
        optim,
        norm_method,
        accum_count,
        accum_steps,
        n_gpu,
        gpu_rank,
        parallel_mode,
        report_manager,
        with_align=(
            True if getattr(config.model.decoder, "lambda_align", 0.0) > 0 else False
        ),  # patch to support non transformer configs
        model_saver=model_saver,
        average_decay=average_decay,
        average_every=average_every,
        earlystopper=earlystopper,
        dropout=dropout,
        attention_dropout=attention_dropout,
        dropout_steps=dropout_steps,
        zero_out_prompt_loss=zero_out_prompt_loss,
        estim_loss_lambda=estim_loss_lambda,
        estim_loss_lambda_steps=estim_loss_lambda_steps,
    )
    return trainer


class Trainer(object):
    """Class that controls the training process.

    Args:
        model(:py:class:`eole.models.model.BaseModel`): model to train
        train_loss(:obj:`eole.utils.loss.LossComputeBase`):
          training loss computation
        valid_loss(:obj:`eole.utils.loss.LossComputeBase`):
          training loss computation
        scoring_preparator(:obj:`eole.predict.utils.ScoringPreparator`):
          preparator for the calculation of metrics via the
          _eval_handler method
        valid_scorers (dict): keeps in memory the current values
          of the validation metrics
        optim(:obj:`eole.utils.optimizers.Optimizer`):
          the optimizer responsible for update
        accum_count(list): accumulate gradients this many times.
        accum_steps(list): steps for accum gradients changes.
        n_gpu (int): number of gpu.
        gpu_rank (int): ordinal rank of the gpu in the list.
        report_manager(:obj:`eole.utils.ReportMgrBase`):
          the object that creates reports, or None
        with_align (bool): whether to jointly lear alignment
          (Transformer)
        model_saver(:obj:`eole.models.ModelSaverBase`): the saver is
          used to save a checkpoint.
          Thus nothing will be saved if this parameter is None.
        average_decay (float): cf opt.average_decay
        average_every (int): average model every x steps.
        earlystopper (:obj:`eole.utils.EarlyStopping`): add early
          stopping mecanism
        dropout (float): dropout value in RNN or FF layers.
        attention_dropout (float): dropaout in attention layers.
        dropout_steps (list): dropout values scheduling in steps.
        zero_out_prompt_loss (bool): whether to zero-out the prompt loss
            (mostly for LLM finetuning).
        estim_loss_lambda (List[float]): weight applied to estimator loss
        estim_loss_lambda_steps (List[int]): steps to apply to estimator values"""

    def __init__(
        self,
        model,
        train_loss,
        valid_loss,
        scoring_preparator,
        valid_scorers,
        optim,
        norm_method="sents",
        accum_count=[1],
        accum_steps=[0],
        n_gpu=1,
        gpu_rank=1,
        parallel_mode="data_parallel",
        report_manager=None,
        with_align=False,
        model_saver=None,
        average_decay=0,
        average_every=1,
        earlystopper=None,
        dropout=[0.3],
        attention_dropout=[0.1],
        dropout_steps=[0],
        zero_out_prompt_loss=False,
        estim_loss_lambda=[1.0],
        estim_loss_lambda_steps=[0],
    ):
        # Basic attributes.

        self.model = model
        self.train_loss = train_loss
        self.estim_loss_lambda_l = estim_loss_lambda
        self.estim_loss_lambda = estim_loss_lambda[0]
        self.estim_loss_lambda_steps = estim_loss_lambda_steps
        self.valid_loss = valid_loss
        self.scoring_preparator = scoring_preparator
        self.valid_scorers = valid_scorers
        self.optim = optim
        self.norm_method = norm_method
        self.accum_count_l = accum_count
        self.accum_count = accum_count[0]
        self.accum_steps = accum_steps
        self.n_gpu = n_gpu
        self.gpu_rank = gpu_rank
        self.parallel_mode = parallel_mode
        self.report_manager = report_manager
        self.with_align = with_align
        self.model_saver = model_saver
        self.average_decay = average_decay
        self.moving_average = None
        self.average_every = average_every
        self.earlystopper = earlystopper
        self.dropout = dropout
        self.attention_dropout = attention_dropout
        self.dropout_steps = dropout_steps
        self.zero_out_prompt_loss = zero_out_prompt_loss

        for i in range(len(self.accum_count_l)):
            assert self.accum_count_l[i] > 0

        # Set model in training mode.
        self.model.train()

    def _eval_handler(self, scorer, preds, texts_ref):
        """Trigger metrics calculations

        Args:
            scorer (:obj:``eole.scorer.Scorer``): scorer.
            preds, texts_ref: outputs of the scorer's `translate` method.

        Returns:
            The metric calculated by the scorer."""

        return scorer.compute_score(preds, texts_ref)

    def _accum_count(self, step):
        for i in range(len(self.accum_steps)):
            if step > self.accum_steps[i]:
                _accum = self.accum_count_l[i]
        return _accum

    def _maybe_update_dropout(self, step):
        for i in range(len(self.dropout_steps)):
            if step > 1 and step == self.dropout_steps[i] + 1:
                self.model.update_dropout(self.dropout[i], self.attention_dropout[i])
                logger.info(
                    "Updated dropout/attn dropout to %f %f at step %d"
                    % (self.dropout[i], self.attention_dropout[i], step)
                )

    def _maybe_update_estim_lambda(self, step):
        for i in range(len(self.estim_loss_lambda_steps)):
            if step > 1 and step == self.estim_loss_lambda_steps[i] + 1:
                self.estim_loss_lambda = self.estim_loss_lambda_l[i]
                logger.info("Updated estimator lambda to %f at step %d" % (self.estim_loss_lambda_l[i], step))

    def _accum_batches(self, iterator):
        batches = []
        normalization = 0
        self.accum_count = self._accum_count(self.optim.training_step)
        for batch, bucket_idx in iterator:
            batches.append(batch)
            if self.norm_method == "tokens":
                num_tokens = batch["tgt"][:, 1:].ne(self.train_loss.padding_idx).sum()
                normalization += num_tokens.item()
                normalization -= len(batch["tgt"])  # don't count for EOS
            else:
                normalization += len(batch["tgt"])
            if len(batches) == self.accum_count:
                yield batches, normalization
                self.accum_count = self._accum_count(self.optim.training_step)
                batches = []
                normalization = 0
        if batches:
            yield batches, normalization

    def _update_average(self, step):
        if self.moving_average is None:
            copy_params = [params.detach() for params in self.model.parameters()]
            self.moving_average = copy_params
        else:
            average_decay = max(self.average_decay, 1 - (step + 1) / (step + 10))
            for (i, avg), cpt in zip(enumerate(self.moving_average), self.model.parameters()):
                self.moving_average[i] = (1 - average_decay) * avg + cpt.detach() * average_decay

    def train(
        self,
        train_iter,
        train_steps,
        save_checkpoint_steps=5000,
        valid_iter=None,
        valid_steps=10000,
    ):
        """The main training loop by iterating over ``train_iter`` and possibly
        running validation on ``valid_iter``.

        Args:
            train_iter: An iterator that returns the next training batch.
            train_steps: Run training for this many iterations.
            save_checkpoint_steps: Save a checkpoint every this many
              iterations.
            valid_iter: A generator that returns the next validation batch.
            valid_steps: Run evaluation every this many iterations.

        Returns:
            :obj:``nmt.Statistics``: training loss statistics"""

        if valid_iter is None:
            logger.info("Start training loop without validation...")
            valid_stats = None
        else:
            logger.info("Start training loop and validate every %d steps...", valid_steps)
        logger.info("Scoring with: {}".format(self.scoring_preparator.transforms))

        total_stats = eole.utils.Statistics()
        report_stats = eole.utils.Statistics()
        self._start_report_manager(start_time=total_stats.start_time)
        # Let's clean the GPUs before training loop
        clear_gpu_cache()

        for i, (batches, normalization) in enumerate(self._accum_batches(train_iter)):

            step = self.optim.training_step
            # UPDATE DROPOUT
            self._maybe_update_dropout(step)
            self._maybe_update_estim_lambda(step)

            if self.n_gpu > 1 and self.parallel_mode == "data_parallel":
                normalization = sum(all_gather_list(normalization))

            self._gradient_accumulation(batches, normalization, total_stats, report_stats)

            if self.average_decay > 0 and i % self.average_every == 0:
                self._update_average(step)

            report_stats = self._maybe_report_training(
                step, train_steps, self.optim.learning_rate(step=step), report_stats
            )

            if valid_iter is not None and step % valid_steps == 0:
                if self.parallel_mode == "tensor_parallel" or self.gpu_rank <= 0:
                    valid_stats = self.validate(valid_iter, moving_average=self.moving_average)

            if step % valid_steps == 0 and self.gpu_rank <= 0:
                self._report_step(
                    self.optim.learning_rate(),
                    step,
                    valid_stats=valid_stats,
                    train_stats=total_stats,
                )

                # Run patience mechanism
                if self.earlystopper is not None:
                    self.earlystopper(valid_stats, step)
                    # If the patience has reached the limit, stop training
                    if self.earlystopper.has_stopped():
                        logger.info("earlystopper has_stopped!")
                        break

            if self.model_saver is not None and (save_checkpoint_steps != 0 and step % save_checkpoint_steps == 0):
                self.model_saver.save(step, moving_average=self.moving_average)

            if train_steps > 0 and step >= train_steps:
                break

        if self.model_saver is not None:
            self.model_saver.save(step, moving_average=self.moving_average)
        return total_stats

    def validate(self, valid_iter, moving_average=None):
        """Validate model.

        Args:
            valid_iter: validate data iterator

        Returns:
            :obj:``nmt.Statistics``: validation loss statistics"""

        valid_model = self.model
        if moving_average:
            # swap model params w/ moving average
            # (and keep the original parameters)
            model_params_data = []
            for avg, param in zip(self.moving_average, valid_model.parameters()):
                model_params_data.append(param.data)
                param.data = avg.data.half() if param.dtype == torch.float16 else avg.data

        # Set model in validating mode.
        valid_model.eval()

        with torch.no_grad():
            stats = eole.utils.Statistics()
            start = time.time()
            for batch, bucket_idx in valid_iter:
                src = batch["src"]
                src_len = batch["srclen"]
                tgt = batch["tgt"]

                with get_autocast(enabled=self.optim.amp):
                    # F-prop through the model.
                    model_out, attns, estim = valid_model(src, tgt, src_len, with_align=self.with_align)

                    # Compute loss.
                    if self.zero_out_prompt_loss:
                        batch = self.valid_loss.ignore_prompt(batch)
                    _, batch_stats, _ = self.valid_loss(batch, model_out, attns, estim=estim)

                    stats.update(batch_stats)
            logger.info(
                """valid stats calculation
                           took: {} s.""".format(
                    time.time() - start
                )
            )

            # Compute validation metrics (at batch.dataset level)
            if len(self.valid_scorers) > 0:
                computed_metrics = {}
                start = time.time()
                with get_autocast(enabled=self.optim.amp):
                    preds, texts_ref = self.scoring_preparator.translate(
                        model=self.model,
                        gpu_rank=self.gpu_rank,
                        step=self.optim.training_step,
                    )
                logger.info(
                    """The translation of the valid dataset for dynamic scoring
                               took : {} s.""".format(
                        time.time() - start
                    )
                )
                for i, metric in enumerate(self.valid_scorers):
                    logger.info("UPDATING VALIDATION {}".format(metric))
                    self.valid_scorers[metric]["value"] = self._eval_handler(
                        scorer=self.valid_scorers[metric]["scorer"],
                        preds=preds,
                        texts_ref=texts_ref,
                    )
                    computed_metrics[metric] = self.valid_scorers[metric]["value"]
                    logger.info("validation {}: {}".format(metric, self.valid_scorers[metric]["value"]))
                    # Compute stats
                    metric_stats = eole.utils.Statistics(0, 0, 0, 0, 0, 0, computed_metrics)

                # Update statistics.
                stats.update(metric_stats)
                valid_model.decoder._disable_cache()

        if moving_average:
            for param_data, param in zip(model_params_data, self.model.parameters()):
                param.data = param_data

        # Set model back to training mode.
        valid_model.train()

        return stats

    def _gradient_accumulation(self, true_batches, normalization, total_stats, report_stats):
        """Function that iterates over big batches = ``true_batches``

        Perform a backward on the loss of each sub_batch and
        finally update the params at the end of the big batch."""

        self.optim.zero_grad(set_to_none=True)

        for k, batch in enumerate(true_batches):

            src = batch["src"]
            src_len = batch["srclen"]
            report_stats.n_src_tokens += src_len.sum().item()
            total_stats.n_src_tokens += src_len.sum().item()
            tgt = batch["tgt"]
            kwargs = {"images": batch["images"], "prefix_len": batch["prefix_len"]}

            try:
                with get_autocast(enabled=self.optim.amp):
                    model_out, attns, estim = self.model(src, tgt, src_len, with_align=self.with_align, **kwargs)
                    if self.zero_out_prompt_loss:
                        # The loss of the prompt will be set to zero.
                        batch = self.train_loss.ignore_prompt(batch)
                    loss, batch_stats, auxloss = self.train_loss(
                        batch,
                        model_out,
                        attns,
                        estim=estim,
                    )
                if loss is not None:
                    loss /= normalization
                    auxloss /= self.accum_count * src_len.size(0)
                    loss = loss + auxloss * self.estim_loss_lambda
                    self.optim.backward(loss)

                total_stats.update(batch_stats)
                report_stats.update(batch_stats)

            except Exception as exc:
                trace_content = traceback.format_exc()
                if "CUDA out of memory" in trace_content:
                    logger.info(
                        "Step %d, cuda OOM - batch removed",
                        self.optim.training_step,
                    )
                    clear_gpu_cache()
                    if self.n_gpu > 1 and self.parallel_mode == "tensor_parallel":
                        torch.distributed.destroy_process_group()
                        sys.exit()
                else:
                    traceback.print_exc()
                    raise exc

        # in case of multi step gradient accumulation,
        # update only after accum batches
        if self.n_gpu > 1 and self.parallel_mode == "data_parallel":
            grads = [p.grad.data for p in self.model.parameters() if p.requires_grad and p.grad is not None]
            all_reduce_and_rescale_tensors(grads, float(self.n_gpu))

        self.optim.step()

    def _start_report_manager(self, start_time=None):
        """Simple function to start report manager (if any)"""

        if self.report_manager is not None:
            if start_time is None:
                self.report_manager.start()
            else:
                self.report_manager.start_time = start_time

    def _maybe_report_training(self, step, num_steps, learning_rate, report_stats):
        """Simple function to report training stats (if report_manager is set)
        see ``eole.utils.ReportManagerBase.report_training`` for doc"""

        if self.report_manager is not None:
            return self.report_manager.report_training(
                step,
                num_steps,
                learning_rate,
                None if self.earlystopper is None else self.earlystopper.current_tolerance,
                report_stats,
                multigpu=self.n_gpu > 1 and self.parallel_mode == "data_parallel",
            )

    def _report_step(self, learning_rate, step, valid_stats=None, train_stats=None):
        """Simple function to report stats (if report_manager is set)
        see ``eole.utils.ReportManagerBase.report_step`` for doc"""

        if self.report_manager is not None:
            return self.report_manager.report_step(
                learning_rate,
                None if self.earlystopper is None else self.earlystopper.current_tolerance,
                step,
                valid_stats=valid_stats,
                train_stats=train_stats,
            )



""" Pytorch Distributed utils
    This piece of code was heavily inspired by the equivalent of Fairseq-py
    https://github.com/pytorch/fairseq
"""

import torch.distributed
from datetime import timedelta
from eole.predict import build_predictor
from eole.transforms import get_transforms_cls, make_transforms
from eole.constants import CorpusTask
from eole.utils.logging import init_logger, logger
from eole.inputters.dynamic_iterator import build_dynamic_dataset_iter


def is_master(config, device_id):
    return config.gpu_ranks[device_id] == 0


def multi_init(config, device_id):
    # config is a running config here
    dist_init_method = "tcp://{master_ip}:{master_port}".format(
        master_ip=config.master_ip, master_port=config.master_port
    )
    dist_world_size = config.world_size
    torch.distributed.init_process_group(
        backend=config.gpu_backend,
        init_method=dist_init_method,
        world_size=dist_world_size,
        rank=config.gpu_ranks[device_id],
        timeout=timedelta(seconds=config.timeout),
    )
    gpu_rank = torch.distributed.get_rank()
    if not is_master(config, device_id):
        logger.disabled = True

    return gpu_rank


def spawned_train(process_fn, config, device_id, error_queue):  # noqa: E501
    """Run `process_fn` on `device_id` with data from `batch_queue`."""
    # config -> full config
    try:
        gpu_rank = multi_init(config.training, device_id)
        if gpu_rank != config.training.gpu_ranks[device_id]:
            raise AssertionError(
                "An error occurred in \
                  Distributed initialization"
            )
        process_fn(config, device_id=device_id)
    except KeyboardInterrupt:
        pass  # killed by parent, do nothing
    except Exception:
        # propagate exception to parent process, keeping original traceback
        import traceback

        error_queue.put((config.training.gpu_ranks[device_id], traceback.format_exc()))


def spawned_infer(config, device_id, error_queue, queue_instruct, queue_result, queue_settings=None):
    """Run various functions for prediction in spawned process on `device_id`."""
    try:
        running_config = config  # will probably switch to config.inference at some point
        gpu_rank = multi_init(running_config, device_id)
        if gpu_rank != running_config.gpu_ranks[device_id]:
            raise AssertionError(
                "An error occurred in \
                  Distributed initialization"
            )
        torch.cuda.set_device(device_id)
        init_logger(config.log_file)
        predictor = build_predictor(config, device_id, logger=logger, report_score=True)
        transforms_cls = get_transforms_cls(config._all_transform)
        transforms = make_transforms(config, transforms_cls, predictor.vocabs)
        while True:
            instruction = queue_instruct.get()
            if queue_settings is not None:
                settings = queue_settings.get()
                predictor.update_settings(**settings)
            if instruction[0] == "stop":
                break
            elif instruction[0] == "infer_list":
                src = instruction[1]
                infer_iter = build_dynamic_dataset_iter(
                    config,
                    transforms,
                    predictor.vocabs,
                    task=CorpusTask.INFER,
                    src=src,
                    device_id=device_id,
                )
                scores, estims, preds = predictor._predict(
                    infer_iter,
                    infer_iter.transforms,
                    config.attn_debug,
                    config.align_debug,
                )
                queue_result.put(scores)
                queue_result.put(estims)
                queue_result.put(preds)
            elif instruction[0] == "infer_file":
                config.src = instruction[1].src
                infer_iter = build_dynamic_dataset_iter(
                    config,
                    transforms,
                    predictor.vocabs,
                    task=CorpusTask.INFER,
                    device_id=device_id,
                )
                scores, estims, preds = predictor._predict(
                    infer_iter,
                    infer_iter.transforms,
                    config.attn_debug,
                    config.align_debug,
                )
                queue_result.put(scores)
                queue_result.put(estims)
                queue_result.put(preds)
            elif instruction[0] == "score_list":
                tgt = instruction[1]
                infer_iter = build_dynamic_dataset_iter(
                    config,
                    transforms,
                    predictor.vocabs,
                    task=CorpusTask.INFER,
                    src=tgt,
                    tgt=tgt,
                    device_id=device_id,
                )
                score_results = predictor._score(infer_iter)
                queue_result.put(score_results)
            elif instruction[0] == "score_file":
                config.src = instruction[1].src
                config.tgt = instruction[1].src
                infer_iter = build_dynamic_dataset_iter(
                    config,
                    transforms,
                    predictor.vocabs,
                    task=CorpusTask.INFER,
                    device_id=device_id,
                )
                score_results = predictor._score(infer_iter)
                queue_result.put(score_results)

    except KeyboardInterrupt:
        pass  # killed by parent, do nothing
    except Exception:
        # propagate exception to parent process, keeping original traceback
        import traceback

        error_queue.put((running_config.gpu_ranks[device_id], traceback.format_exc()))



"""Define constant values used across the project."""

from enum import Enum
import torch
import torch.nn.functional as F
from functools import partial


class DefaultTokens(object):
    PAD = "<blank>"
    BOS = "<s>"
    EOS = "</s>"
    UNK = "<unk>"
    MASK = "<mask>"
    VOCAB_PAD = "averyunlikelytoken"
    SENT_FULL_STOPS = [".", "?", "!"]
    PHRASE_TABLE_SEPARATOR = "|||"
    ALIGNMENT_SEPARATOR = " ||| "
    SEP = "｟newline｠"
    MASK_BEFORE = "｟_mask_before_｠"


class CorpusName(str, Enum):
    VALID = "valid"
    TRAIN = "train"
    SAMPLE = "sample"
    INFER = "infer"


class CorpusTask(str, Enum):
    TRAIN = "train"
    VALID = "valid"
    INFER = "infer"


class SubwordMarker(str, Enum):
    SPACER = "▁"
    JOINER = "￭"
    BEGIN_UPPERCASE = "｟mrk_begin_case_region_U｠"
    END_UPPERCASE = "｟mrk_end_case_region_U｠"
    BEGIN_CASED = "｟mrk_case_modifier_C｠"


class ModelType(str, Enum):
    DECODER = "decoder"
    ENCODER_DECODER = "encoder_decoder"
    ENCODER = "encoder"


class PositionEncodingType(str, Enum):
    SinusoidalInterleaved = "SinusoidalInterleaved"
    SinusoidalConcat = "SinusoidalConcat"
    Learned = "Learned"
    Relative = "Relative"
    Rotary = "Rotary"
    Alibi = "Alibi"


def fused_gated_gelu(x):
    d = x.shape[-1] // 2
    output_shape = x.shape[:-1] + (d,)
    out = torch.empty(output_shape, dtype=x.dtype, device=x.device)
    torch.ops._C.gelu_and_mul(out, x)
    return out


def fused_gated_silu(x):
    d = x.shape[-1] // 2
    # basically doing return F.silu(x[..., :d]) * x[..., d:]
    output_shape = x.shape[:-1] + (d,)
    out = torch.empty(output_shape, dtype=x.dtype, device=x.device)
    torch.ops._C.silu_and_mul(out, x)
    return out


def fused_gated_gelu_tanh(x):
    d = x.shape[-1] // 2
    output_shape = x.shape[:-1] + (d,)
    out = torch.empty(output_shape, dtype=x.dtype, device=x.device)
    torch.ops._C.gelu_tanh_and_mul(out, x)
    return out


@torch.jit.script
def quick_gelu(x):
    return x * torch.sigmoid(1.702 * x)


class ActivationFunction(str, Enum):
    relu = "relu"
    gelu = "gelu"
    gelu_tanh = "gelu-tanh"
    quick_gelu = "quick_gelu"
    silu = "silu"
    gated_gelu = "gated-gelu"
    gated_gelu_tanh = "gated-gelu-tanh"
    gated_silu = "gated-silu"
    fused_gated_gelu = "fused-gated-gelu"
    fused_gated_gelu_tanh = "fused-gated-gelu-tanh"
    fused_gated_silu = "fused-gated-silu"


class TransformType(str, Enum):
    Default = "any"
    Train = "train"
    Predict = "predict"


ACTIVATION_FUNCTIONS = {
    ActivationFunction.relu: F.relu,
    ActivationFunction.gelu: F.gelu,
    ActivationFunction.silu: F.silu,
    ActivationFunction.quick_gelu: quick_gelu,
    ActivationFunction.gated_gelu: F.gelu,
    ActivationFunction.fused_gated_gelu: fused_gated_gelu,
    ActivationFunction.gated_silu: F.silu,
    ActivationFunction.fused_gated_silu: fused_gated_silu,
    ActivationFunction.gelu_tanh: partial(F.gelu, approximate="tanh"),
    ActivationFunction.gated_gelu_tanh: partial(F.gelu, approximate="tanh"),
    ActivationFunction.fused_gated_gelu_tanh: fused_gated_gelu_tanh,
}


class LayerNormFP32(torch.nn.LayerNorm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, input):
        orig_dtype = input.dtype  # Save original dtype (likely bfloat16)
        input = input.to(torch.float32)  # Process in float32
        output = super().forward(input)  # Apply LayerNorm
        return output.to(orig_dtype)  # Convert back to original dtype


class LazyLayerNormDict:
    """Dictionary-like object that lazily imports heavy LayerNorm classes."""

    def __init__(self):
        self._registry = {
            "standardFP32": LayerNormFP32,
            "standard": torch.nn.LayerNorm,
            "rms": None,  # Placeholder
            "gemma-rms": None,  # Placeholder
        }

    def __getitem__(self, key):
        if key not in self._registry:
            raise KeyError(f"Unknown LayerNorm type: {key}")

        # Lazy import only when first accessed
        if key == "rms" and self._registry[key] is None:
            from eole.modules.rmsnorm import RMSNorm

            self._registry[key] = RMSNorm
        elif key == "gemma-rms" and self._registry[key] is None:
            from eole.modules.rmsnorm import GemmaRMSNorm

            self._registry[key] = GemmaRMSNorm

        return self._registry[key]


LayerNorm = LazyLayerNormDict()


TORCH_DTYPES = {
    "fp32": torch.float32,
    "fp16": torch.float16,
    "bf16": torch.bfloat16,
    "int8": torch.int8,
    "torch.float32": torch.float32,
    "torch.float16": torch.float16,
    "torch.bfloat16": torch.bfloat16,
    "torch.int8": torch.int8,
    "float32": torch.float32,
    "float16": torch.float16,
    "bfloat16": torch.bfloat16,
}



"""Tokenization transforms that also numericalize, such as HF, or Mistral"""

import os
from eole.utils.logging import logger
from eole.transforms import register_transform
from .transform import Transform, TransformConfig  # , ObservableStats
from pydantic import Field
from eole.constants import DefaultTokens


class BaseIdTokenizerConfig(TransformConfig):
    # model_name: str | None = Field(default=None)  # if tokenizer deduced from model
    path: str | None = Field(default=None)  # if local tokenizer


class HuggingfaceTokenizerConfig(BaseIdTokenizerConfig):
    huggingface_model: str | None = Field(default=None)  # if retrieved from huggingface
    max_length: int | None = Field(default=None)


class IntTokenizerTransform(Transform):
    """IntTokenizer transform abstract class."""

    config_model = BaseIdTokenizerConfig
    output_type = "ids"

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        # self.model_name = self.config.model_name
        self.path = self.config.path

    def warm_up(self, vocabs=None):
        # TODO: factorize here
        raise NotImplementedError


@register_transform(name="huggingface_tokenize")
class HuggingfaceTokenizer(IntTokenizerTransform):
    config_model = HuggingfaceTokenizerConfig

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        super()._parse_config()
        self.huggingface_model = self.config.huggingface_model
        self.max_length = self.config.max_length

    def warm_up(self, vocabs=None):
        from tokenizers.processors import TemplateProcessing

        if self.huggingface_model is not None:
            from transformers import AutoTokenizer

            self.tokenizers = {}

            self.tokenizers["src"] = AutoTokenizer.from_pretrained(self.huggingface_model, legacy=False)
            # TODO: this needs to be tested and adapted for various models
            self.tokenizers["tgt"] = AutoTokenizer.from_pretrained(self.huggingface_model, legacy=False)
            logger.info(f"Initialized tokenizers from HF model: {self.huggingface_model}")
        elif self.path is not None:
            if os.path.exists(self.path):
                from tokenizers import Tokenizer

                self.tokenizers["src"] = Tokenizer.from_file(self.path, legacy=False)
                # TODO: might be more efficient to have a single tokenizer,
                # and apply the eos logic on top
                self.tokenizers["tgt"] = Tokenizer.from_file(self.path, legacy=False)
            else:
                raise FileNotFoundError(self.path)
            logger.info(f"Initialized tokenizers from local file: {self.path}")
        else:
            raise RuntimeError(f"Either model_name or path must be configured for {self.name} transform")
        # https://github.com/huggingface/transformers/issues/22794#issuecomment-2092623992
        # bos = self.tokenizers["tgt"].bos_token
        # bos_id = self.tokenizers["tgt"].bos_token_id
        eos = self.tokenizers["tgt"].eos_token
        eos_id = self.tokenizers["tgt"].eos_token_id
        self.tokenizers["tgt"]._tokenizer.post_processor = TemplateProcessing(
            single=f"$A:0 {eos}:0",
            pair=f"$A:0 {eos}:0 $B:1 {eos}:1",
            special_tokens=[
                # (f"{bos}", bos_id),
                (f"{eos}", eos_id)
            ],
        )

    def tokenize_string(self, string, side="src", is_train=False):
        if self.max_length is not None and is_train:
            max_length = self.max_length
            decoder_start_token = self.full_config.decoder_start_token
            if side == "tgt" and self.full_config.model.encoder is None and decoder_start_token == "":
                max_length += 1
            kwargs = {"max_length": max_length, "truncation": True}
        else:
            kwargs = {}
        tokens = self.tokenizers[side].encode(string.replace(DefaultTokens.SEP, "\n"), **kwargs)
        return tokens

    def apply(self, example, is_train=False, stats=None, **kwargs):
        if isinstance(example["src"], str):
            src_tokens = self.tokenize_string(example["src"], side="src", is_train=is_train)
        elif isinstance(example["src"], list):
            src_tokens = self.tokenize_string(" ".join(example["src"]), side="src", is_train=is_train)
        else:
            raise ValueError(f"Unsupported src type: {type(example['src'])}")
        example["src_ids"] = src_tokens
        if example.get("tgt", None) is not None:
            if isinstance(example["tgt"], str):
                tgt_tokens = self.tokenize_string(example["tgt"], side="tgt", is_train=is_train)
            elif isinstance(example["tgt"], list):
                tgt_tokens = self.tokenize_string(" ".join(example["tgt"]), side="tgt", is_train=is_train)
            example["tgt_ids"] = tgt_tokens
        else:  # tricky case: when using tokenize_id and tgt is None then we use src instead (LM classic training)
            if isinstance(example["src"], str):
                tgt_tokens = self.tokenize_string(example["src"], side="tgt", is_train=is_train)
            elif isinstance(example["src"], list):
                tgt_tokens = self.tokenize_string(" ".join(example["src"]), side="tgt", is_train=is_train)
            example["tgt_ids"] = tgt_tokens
        return example

    def apply_reverse(self, predicted):
        detokenized = self.tokenizers["tgt"].decode(predicted).replace("\n", DefaultTokens.SEP)
        return detokenized



from eole.transforms import register_transform
from .transform import Transform, TransformConfig
from pydantic import Field
import unicodedata
import random


class UpperCaseConfig(TransformConfig):
    upper_corpus_ratio: float | None = Field(default=0.01, description="Corpus ratio to apply uppercasing.")


@register_transform(name="uppercase")
class UpperCaseTransform(Transform):
    """
    Convert source and target examples to uppercase.

    This transform uses `unicodedata` to normalize the converted
    uppercase strings as this is needed for some languages (e.g. Greek).
    One issue is that the normalization removes all diacritics and
    accents from the uppercased strings, even though in few occasions some
    diacritics should be kept even in the uppercased form.
    """

    config_model = UpperCaseConfig

    def __init__(self, config):
        super().__init__(config)

    def _parse_config(self):
        self.upper_corpus_ratio = self.config.upper_corpus_ratio

    def apply(self, example, is_train=False, stats=None, **kwargs):
        """Convert source and target examples to uppercase."""

        if random.random() > self.upper_corpus_ratio:
            return example

        src_str = " ".join(example["src"])
        src_str = "".join(c for c in unicodedata.normalize("NFD", src_str.upper()) if unicodedata.category(c) != "Mn")
        example["src"] = src_str.split(" ")

        if example["tgt"] is not None:
            tgt_str = " ".join(example["tgt"])
            tgt_str = "".join(
                c for c in unicodedata.normalize("NFD", tgt_str.upper()) if unicodedata.category(c) != "Mn"
            )
            example["tgt"] = tgt_str.split(" ")

        return example



#!/usr/bin/env python
import json
import torch
import pyonmttok
from eole.inputters.inputter import vocabs_to_dict
from sentencepiece import SentencePieceProcessor
import os
from safetensors.torch import save_file

# from transformers import T5ForConditionalGeneration

from eole.config.models import (
    EmbeddingsConfig,
    TransformerModelConfig,
)
from eole.config.run import TrainConfig
from eole.config.training import TrainingConfig
from eole.config import recursive_model_fields_set
from eole.bin import BaseBin, register_bin


"""
Special Note to T5 models conversion.

T5 does not rescale MHE query (by dividing by square_root of dim_per_head)
query /= math.sqrt(self.dim_per_head)

So all query weights are rescaled directly in the checkpoint so that
no change is needed in the legacy MHA code.
"""


class Tokenizer:
    def __init__(self, model_path: str):
        assert os.path.isfile(model_path), model_path
        self.sp_model = SentencePieceProcessor(model_file=model_path)
        self.n_words: int = self.sp_model.vocab_size()
        self.bos_id: int = self.sp_model.bos_id()
        self.eos_id: int = self.sp_model.eos_id()
        self.pad_id: int = self.sp_model.pad_id()
        assert self.sp_model.vocab_size() == self.sp_model.get_piece_size()
        self.vocab = [self.sp_model.id_to_piece(i) for i in range(self.n_words)]


@register_bin(name="T5")
class T5Converter(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "--model_dir",
            type=str,
            required=True,
            help="""Path to the model directory""",
        )
        parser.add_argument(
            "--tokenizer_model",
            type=str,
            required=True,
            help="""Path to the tokenizer model""",
        )
        parser.add_argument("--output", type=str, required=True, help="""Path to the model directory""")
        parser.add_argument("--nshards", type=int, default=1, help="""Path to the model directory""")

    @classmethod
    def run(cls, args):
        import eole.utils.sentencepiece_model_pb2 as spmodel

        # load this from pytorch_model.bin directly instead
        # model = T5ForConditionalGeneration.from_pretrained(
        #     args.model_dir,
        #     torch_dtype=torch.float16,
        #     device_map={"": "cpu"},
        #     trust_remote_code=True,
        # )
        # checkpoint = model.state_dict()
        checkpoint = torch.load(os.path.join(args.model_dir, "pytorch_model.bin"), map_location="cpu")
        checkpoint = {k: v.half() for k, v in checkpoint.items()}

        params_json = os.path.join(args.model_dir, "config.json")
        with open(params_json, encoding="utf-8") as fparam:
            params = json.load(fparam)

        decoder_layers = params["num_decoder_layers"]
        encoder_layers = params["num_layers"]
        src_word_vec_size = params["d_model"]
        tgt_word_vec_size = params["d_model"]
        hidden_size = params["d_model"]
        heads = params["num_heads"]
        vocab_size = params["vocab_size"]
        transformer_ff = params["d_ff"]
        dimperhead = hidden_size // heads

        for shard in range(args.nshards):

            print("starting output shard: %d/%d" % (shard + 1, args.nshards))
            eole_safetensor = {}

            if shard == 0:
                eole_safetensor["tgt_emb.embeddings.weight"] = checkpoint["decoder.embed_tokens.weight"].to(
                    torch.float16
                )
                eole_safetensor["decoder.layer_norm.weight"] = checkpoint["decoder.final_layer_norm.weight"].to(
                    torch.float16
                )
                eole_safetensor["src_emb.embeddings.weight"] = checkpoint["encoder.embed_tokens.weight"].to(
                    torch.float16
                )
                eole_safetensor["encoder.layer_norm.weight"] = checkpoint["encoder.final_layer_norm.weight"].to(
                    torch.float16
                )
                eole_safetensor["generator.weight"] = checkpoint["lm_head.weight"].to(torch.float16)
                eole_safetensor["generator.bias"] = torch.zeros(
                    eole_safetensor["generator.weight"].size(0), dtype=torch.float16
                )

            for i in range(
                -(decoder_layers // -args.nshards) * shard,
                min(-(decoder_layers // -args.nshards) * (shard + 1), decoder_layers),
                1,
            ):
                eole_safetensor["encoder.transformer_layers." + str(i) + ".self_attn.linear_query.weight"] = (
                    checkpoint["encoder.block." + str(i) + ".layer.0.SelfAttention.q.weight"] / (dimperhead**-0.5)
                ).to(torch.float16)
                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_query.weight"] = (
                    checkpoint["decoder.block." + str(i) + ".layer.0.SelfAttention.q.weight"] / (dimperhead**-0.5)
                ).to(torch.float16)
                eole_safetensor["encoder.transformer_layers." + str(i) + ".self_attn.linear_keys.weight"] = checkpoint[
                    "encoder.block." + str(i) + ".layer.0.SelfAttention.k.weight"
                ].to(torch.float16)
                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_keys.weight"] = checkpoint[
                    "decoder.block." + str(i) + ".layer.0.SelfAttention.k.weight"
                ].to(torch.float16)
                eole_safetensor["encoder.transformer_layers." + str(i) + ".self_attn.linear_values.weight"] = (
                    checkpoint["encoder.block." + str(i) + ".layer.0.SelfAttention.v.weight"].to(torch.float16)
                )
                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.linear_values.weight"] = (
                    checkpoint["decoder.block." + str(i) + ".layer.0.SelfAttention.v.weight"].to(torch.float16)
                )

                eole_safetensor["encoder.transformer_layers." + str(i) + ".self_attn.final_linear.weight"] = checkpoint[
                    "encoder.block." + str(i) + ".layer.0.SelfAttention.o.weight"
                ].to(torch.float16)
                eole_safetensor["decoder.transformer_layers." + str(i) + ".self_attn.final_linear.weight"] = checkpoint[
                    "decoder.block." + str(i) + ".layer.0.SelfAttention.o.weight"
                ].to(torch.float16)

                eole_safetensor[
                    "encoder.transformer_layers." + str(i) + ".self_attn.relative_attention_bias.weight"
                ] = checkpoint["encoder.block.0.layer.0.SelfAttention.relative_attention_bias.weight"].to(torch.float16)
                eole_safetensor[
                    "decoder.transformer_layers." + str(i) + ".self_attn.relative_attention_bias.weight"
                ] = checkpoint["decoder.block.0.layer.0.SelfAttention.relative_attention_bias.weight"].to(torch.float16)

                eole_safetensor["encoder.transformer_layers." + str(i) + ".layer_norm.weight"] = checkpoint[
                    "encoder.block." + str(i) + ".layer.0.layer_norm.weight"
                ].to(torch.float16)
                eole_safetensor["decoder.transformer_layers." + str(i) + ".layer_norm_1.weight"] = checkpoint[
                    "decoder.block." + str(i) + ".layer.0.layer_norm.weight"
                ].to(torch.float16)

                eole_safetensor["encoder.transformer_layers." + str(i) + ".feed_forward.w_1.weight"] = checkpoint[
                    "encoder.block." + str(i) + ".layer.1.DenseReluDense.wi_0.weight"
                ].to(torch.float16)

                eole_safetensor["encoder.transformer_layers." + str(i) + ".feed_forward.w_2.weight"] = checkpoint[
                    "encoder.block." + str(i) + ".layer.1.DenseReluDense.wo.weight"
                ].to(torch.float16)
                eole_safetensor["encoder.transformer_layers." + str(i) + ".feed_forward.w_3.weight"] = checkpoint[
                    "encoder.block." + str(i) + ".layer.1.DenseReluDense.wi_1.weight"
                ].to(torch.float16)

                eole_safetensor["decoder.transformer_layers." + str(i) + ".context_attn.linear_query.weight"] = (
                    checkpoint["decoder.block." + str(i) + ".layer.1.EncDecAttention.q.weight"] / (dimperhead**-0.5)
                ).to(torch.float16)
                eole_safetensor["decoder.transformer_layers." + str(i) + ".context_attn.linear_keys.weight"] = (
                    checkpoint["decoder.block." + str(i) + ".layer.1.EncDecAttention.k.weight"].to(torch.float16)
                )
                eole_safetensor["decoder.transformer_layers." + str(i) + ".context_attn.linear_values.weight"] = (
                    checkpoint["decoder.block." + str(i) + ".layer.1.EncDecAttention.v.weight"].to(torch.float16)
                )
                eole_safetensor["decoder.transformer_layers." + str(i) + ".context_attn.final_linear.weight"] = (
                    checkpoint["decoder.block." + str(i) + ".layer.1.EncDecAttention.o.weight"].to(torch.float16)
                )

                eole_safetensor["decoder.transformer_layers." + str(i) + ".layer_norm_2.weight"] = checkpoint[
                    "decoder.block." + str(i) + ".layer.1.layer_norm.weight"
                ].to(torch.float16)

                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.w_1.weight"] = checkpoint[
                    "decoder.block." + str(i) + ".layer.2.DenseReluDense.wi_0.weight"
                ].to(torch.float16)

                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.w_2.weight"] = checkpoint[
                    "decoder.block." + str(i) + ".layer.2.DenseReluDense.wo.weight"
                ].to(torch.float16)
                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.w_3.weight"] = checkpoint[
                    "decoder.block." + str(i) + ".layer.2.DenseReluDense.wi_1.weight"
                ].to(torch.float16)

                eole_safetensor["encoder.transformer_layers." + str(i) + ".feed_forward.layer_norm.weight"] = (
                    checkpoint["encoder.block." + str(i) + ".layer.1.layer_norm.weight"].to(torch.float16)
                )
                eole_safetensor["decoder.transformer_layers." + str(i) + ".feed_forward.layer_norm.weight"] = (
                    checkpoint["decoder.block." + str(i) + ".layer.2.layer_norm.weight"].to(torch.float16)
                )

            if shard == 0:
                transformer_ff = eole_safetensor["decoder.transformer_layers.0.feed_forward.w_1.weight"].size(0)
                vocab_size = eole_safetensor["generator.weight"].size(0)
            print("Saving output model shard: %d" % shard)
            save_file(
                eole_safetensor,
                os.path.join(args.output, "model.{:02d}.safetensors".format(shard)),
            )

        tokenizer = Tokenizer(model_path=args.tokenizer_model)
        vocabs = {}
        vocab = tokenizer.vocab

        src_vocab = pyonmttok.build_vocab_from_tokens(
            vocab,
            maximum_size=tokenizer.n_words,
            special_tokens=["<blank>", "</s>", "<unk>"],
        )
        vocabs["src"] = src_vocab
        vocabs["tgt"] = src_vocab

        vocabs["decoder_start_token"] = "<blank>"

        vocab_dict = vocabs_to_dict(vocabs)
        with open(os.path.join(args.output, "vocab.json"), "w", encoding="utf-8") as f:
            json.dump(vocab_dict, f, indent=2, ensure_ascii=False)

        with open(os.path.join(args.output, "vocab.txt"), "w", encoding="utf-8") as vocabfile:
            for tok in vocab_dict["src"]:
                vocabfile.write(tok + "\n")

        position_encoding = {
            "position_encoding_type": "Relative",
            "n_positions": params["relative_attention_max_distance"],
        }

        config = TrainConfig(
            data=None,
            skip_empty_level="silent",  # default is "warning"
            save_data=None,
            n_sample=0,
            src_vocab=None,
            tgt_vocab=None,
            share_vocab=True,
            src_vocab_size=vocab_size,
            tgt_vocab_size=vocab_size,
            vocab_size_multiple=8,
            decoder_start_token=vocabs["decoder_start_token"],
            transforms=["filtertoolong"],
            transforms_configs={"filtertoolong": {"src_seq_length": 512, "tgt_seq_length": 512}},
            model=TransformerModelConfig(
                layers=decoder_layers,
                hidden_size=hidden_size,
                heads=heads,
                transformer_ff=transformer_ff,
                embeddings=EmbeddingsConfig(
                    src_word_vec_size=src_word_vec_size,
                    tgt_word_vec_size=tgt_word_vec_size,
                    **position_encoding,
                ),
                layer_norm="rms",
                pos_ffn_activation_fn="gated-gelu",
                relative_positions_buckets=params["relative_attention_num_buckets"],
                parallel_residual=False,
                add_qkvbias=False,
                add_ffnbias=False,
                encoder={"layers": encoder_layers},
                decoder={"layers": decoder_layers},
            ),
            training=TrainingConfig(
                compute_dtype="fp16",
                batch_size=896,
                batch_size_multiple=1,
                batch_type="tokens",
                normalization="tokens",
                accum_count=[32],
                accum_steps=[0],
                valid_batch_size=256,
            ),
        )

        config_dict = recursive_model_fields_set(config)
        with open(os.path.join(args.output, "config.json"), "w", encoding="utf-8") as f:
            json.dump(config_dict, f, indent=2, ensure_ascii=False)

        serializedStr = open(args.tokenizer_model, "rb").read()
        m = spmodel.ModelProto()
        m.ParseFromString(serializedStr)
        m.pieces[0].piece = "<blank>"
        with open(args.tokenizer_model + ".new", "wb") as f:
            f.write(m.SerializeToString())

        print(
            "The sentencepiece model %s has been patched with <blank> instead of <pad>",
            args.tokenizer_model,
        )
        print("With eole use the model with the .new extension or rename it")



"""
Implementation of "Attention is All You Need" Transformer Decoder
"""

import torch
import torch.nn as nn
from eole.decoders.decoder import DecoderBase
from eole.modules.multi_headed_attn import SelfMHA, ContextMHA
from eole.modules.transformer_mlp import MLP
from eole.modules.moe import MoE
from eole.modules.rope import build_rope
from eole.constants import LayerNorm


class TransformerDecoderLayer(nn.Module):
    """Single layer of a transformer decoder

    Args:
        decoder_config (eole.config.TransformerEncoderConfig): full encoder config
        running_config (TrainingConfig / InferenceConfig)
        with_cross_attn (True when used with an encoder)
    """

    def __init__(self, decoder_config, idx: int, running_config=None, with_cross_attn=False):
        super(TransformerDecoderLayer, self).__init__()
        self.parallel_residual = decoder_config.parallel_residual
        self.shared_layer_norm = decoder_config.shared_layer_norm
        self.dropout_p = getattr(running_config, "dropout", [0.0])[0]
        self.full_context_alignment = decoder_config.full_context_alignment
        self.alignment_heads = decoder_config.alignment_heads
        self.ffn_layernorm = decoder_config.ffn_layernorm

        # order of layers corresponds to forward flow of tensors
        self.input_layernorm = LayerNorm[decoder_config.layer_norm](
            decoder_config.hidden_size, eps=decoder_config.norm_eps
        )
        self.self_attn = SelfMHA(
            decoder_config,
            running_config=running_config,
        )
        self.dropout = nn.Dropout(self.dropout_p)
        if with_cross_attn:
            self.precontext_layernorm = LayerNorm[decoder_config.layer_norm](
                decoder_config.hidden_size, eps=decoder_config.norm_eps
            )
            self.context_attn = ContextMHA(
                decoder_config,
                running_config=running_config,
            )
        else:
            self.context_attn = None

        if self.ffn_layernorm:
            self.pre_feedforward_layernorm = LayerNorm[decoder_config.layer_norm](
                decoder_config.hidden_size, eps=decoder_config.norm_eps
            )
            self.post_feedforward_layernorm = LayerNorm[decoder_config.layer_norm](
                decoder_config.hidden_size, eps=decoder_config.norm_eps
            )

        if decoder_config.parallel_residual and not decoder_config.shared_layer_norm:
            self.residual_layernorm = LayerNorm[decoder_config.layer_norm](
                decoder_config.hidden_size, eps=decoder_config.norm_eps
            )
        if self.ffn_layernorm or not self.parallel_residual:
            self.post_attention_layernorm = LayerNorm[decoder_config.layer_norm](
                decoder_config.hidden_size, eps=decoder_config.norm_eps
            )
        if decoder_config.num_experts > 0:
            if idx >= decoder_config.first_k_dense_replace:
                self.mlp = MoE(decoder_config, running_config)
            else:
                self.mlp = MLP(
                    decoder_config,
                    running_config=running_config,
                )
        else:
            self.mlp = MLP(
                decoder_config,
                running_config=running_config,
            )

    def _mlp(self, hidden_states):
        if self.ffn_layernorm:
            hidden_states = self.pre_feedforward_layernorm(hidden_states)
            hidden_states = self.mlp(hidden_states)
            hidden_states = self.post_feedforward_layernorm(hidden_states)
        else:
            hidden_states = self.mlp(hidden_states)
        return hidden_states

    def forward(self, layer_in, **kwargs):
        """
        Args:
            layer_in (FloatTensor): ``(batch_size, tgt_len, model_dim)``
            **kwargs
                enc_out: (only when using encoder/decoder)
                src_pad_mask: (only when using encoder/decoder)
                attn_mask (BoolTensor): ``(batch_size, 1, src/tgt_len, tgt_len)``
                step: int
                return_attention: bool
                position_embeddings (FloatTensor): rotary position encodings, if any

        Returns:
            (FloatTensor, FloatTensor):
            * layer_out ``(batch_size, tgt_len, model_dim)``
            * attns
        """

        enc_out = kwargs.pop("enc_out", None)
        attn_mask = kwargs.pop("attn_mask", None)
        src_pad_mask = kwargs.pop("src_pad_mask", None)
        step = kwargs.pop("step", None)
        return_attn = kwargs.pop("return_attn", False)
        position_embeddings = kwargs.pop("position_embeddings", None)

        norm_layer_in = self.input_layernorm(layer_in)

        self_attn, attns = self.self_attn(
            norm_layer_in,
            attn_mask=attn_mask,
            step=step,
            return_attn=return_attn,
            position_embeddings=position_embeddings,
        )

        if self.dropout_p > 0:
            self_attn = self.dropout(self_attn)

        if self.ffn_layernorm:
            # NOTE: this case was added for Gemma2 support and might be extended for further ctx_attn support
            ff_in = layer_in + self.post_attention_layernorm(self_attn)
            layer_out = ff_in + self._mlp(ff_in)
            return layer_out, attns
        else:
            self_attn.add_(layer_in)

        if self.parallel_residual:
            if self.context_attn:
                ctx_attn, attns = self.context_attn(
                    enc_out,
                    enc_out,
                    norm_layer_in,
                    attn_mask=~src_pad_mask,
                    return_attn=return_attn,
                )
                ctx_attn.add_(self_attn)
            else:
                ctx_attn = self_attn
            if not self.shared_layer_norm:
                ff_in = self.residual_layernorm(layer_in)
            else:
                ff_in = norm_layer_in
        else:
            if self.context_attn:
                norm_query = self.precontext_layernorm(self_attn + layer_in)
                ctx_attn, attns = self.context_attn(
                    enc_out,
                    enc_out,
                    norm_query,
                    attn_mask=~src_pad_mask,
                    return_attn=return_attn,
                )
                if self.dropout_p > 0:
                    ctx_attn = self.dropout(ctx_attn)
                ctx_attn.add_(self_attn)
            else:
                ctx_attn = self_attn
            ff_in = self.post_attention_layernorm(ctx_attn)
        # we apply residual with un-normed
        return ctx_attn + self.mlp(ff_in), attns

    def get_attn_align(self, layer_in, **kwargs):
        """:cite:`garg2019jointly`."""
        attns = kwargs.pop("attns", None)
        if self.full_context_alignment:
            # return _, (B, Q_len, K_len)
            _, attns = self.forward(layer_in, **kwargs)
        if self.alignment_heads > 0:
            attns = attns[:, : self.alignment_heads, :, :].contiguous()
        # layer average attention across heads, get ``(B, Q, K)``
        # Case 1: no full_context, no align heads -> layer avg baseline
        # Case 2: no full_context, 1 align heads -> guided align
        # Case 3: full_context, 1 align heads -> full cte guided align
        attn_align = attns.mean(dim=1)

        return attn_align

    def update_dropout(self, dropout, attention_dropout):
        self.self_attn.update_dropout(attention_dropout)
        if self.context_attn:
            self.context_attn.update_dropout(attention_dropout)
        self.mlp.update_dropout(dropout)
        self.dropout.p = dropout


class TransformerDecoder(DecoderBase):
    """The Transformer encoder from "Attention is All You Need"
    :cite:`DBLP:journals/corr/VaswaniSPUJGKP17`

    Args:
        decoder_config (eole.config.TransformerEncoderConfig): full encoder config
        running_config (TrainingConfig / InferenceConfig)
        with_cross_attn (True when used with an encoder)
    """

    def __init__(
        self,
        decoder_config,
        running_config=None,
        with_cross_attn=False,
    ):
        super(TransformerDecoder, self).__init__()
        self.alignment_layer = decoder_config.alignment_layer
        self.with_cross_attn = with_cross_attn
        self.sliding_window = decoder_config.sliding_window
        self.rope = build_rope(decoder_config)
        if hasattr(decoder_config, "rope_config") and getattr(decoder_config.rope_config, "interleave_local", 0) > 0:
            self.rope_local = build_rope(decoder_config, variant="local")
        else:
            self.rope_local = None
        self.interleave_local = getattr(decoder_config.rope_config, "interleave_local", 0) or 1
        self.transformer_layers = nn.ModuleList(
            [
                TransformerDecoderLayer(
                    decoder_config,
                    running_config=running_config,
                    with_cross_attn=with_cross_attn,
                    idx=i,
                )
                for i in range(decoder_config.layers)
            ]
        )
        self.layer_norm = LayerNorm[decoder_config.layer_norm](decoder_config.hidden_size, eps=decoder_config.norm_eps)
        self.LM_type = getattr(decoder_config, "LM_type", "causal")
        self._disable_cache()

    @classmethod
    def from_config(cls, decoder_config, running_config=None, with_cross_attn=False):
        """Alternate constructor."""
        return cls(
            decoder_config,
            running_config=running_config,
            with_cross_attn=with_cross_attn,
        )

    def init_state(self, **kwargs):
        """Initialize decoder state."""
        pass

    def map_state(self, fn):
        if self.left_pad_mask is not None:
            self.left_pad_mask = fn(self.left_pad_mask)
        for layer in self.transformer_layers:
            if self.with_cross_attn:
                if layer.context_attn.kcache is not None:
                    layer.context_attn.kcache = fn(layer.context_attn.kcache)
                    layer.context_attn.vcache = fn(layer.context_attn.vcache)
            if layer.self_attn.kcache is not None:
                layer.self_attn.kcache = fn(layer.self_attn.kcache)
                layer.self_attn.vcache = fn(layer.self_attn.vcache)

    def update_dropout(self, dropout, attention_dropout):
        for layer in self.transformer_layers:
            layer.update_dropout(dropout, attention_dropout)

    def _causal_attn_mask(self, tgt_pad_mask, prefix_len=None):
        batch_size, _, tgt_len = tgt_pad_mask.size()
        device = tgt_pad_mask.device
        # Add triangular future_mask and pad_mask, result mask in (B, T, T).
        future_mask = torch.tril(
            torch.ones((tgt_len, tgt_len), device=tgt_pad_mask.device, dtype=torch.bool),
            diagonal=0,
        )  # Shape: (tgt_len, tgt_len)
        if self.sliding_window > 0:
            future_mask = future_mask.triu_(-self.sliding_window)
        if self.LM_type == "prefix" and prefix_len is not None:
            idx = torch.arange(tgt_len, device=device)
            block = (idx.view(1, tgt_len, 1) < prefix_len.view(batch_size, 1, 1)) & (
                idx.view(1, 1, tgt_len) < prefix_len.view(batch_size, 1, 1)
            )
            future_mask = torch.where(block, torch.ones_like(future_mask), future_mask)  # (B, T, T)
            attn_mask = ~(tgt_pad_mask | tgt_pad_mask.transpose(2, 1)) & future_mask
        else:
            future_mask = future_mask.unsqueeze(0)
            attn_mask = ~tgt_pad_mask & future_mask
        return attn_mask.unsqueeze(1)  # (batch x 1 x 1 x tgt_len)

    def _update_causal_mask(self, attn_mask, image_locations):
        # replicating HF code, can probably be simplified
        token_type_ids = torch.where(image_locations, torch.tensor(1), torch.tensor(0))
        token_type_mask = token_type_ids.unsqueeze(1) == token_type_ids.unsqueeze(2)
        token_type_mask[token_type_ids == 0] = False
        token_type_mask = token_type_mask.unsqueeze(1).to(attn_mask.device, dtype=torch.bool)
        attn_mask = attn_mask.masked_fill(token_type_mask, True)
        return attn_mask

    def forward(self, emb, **kwargs):
        """Decode, possibly stepwise.

        Args:
            emb (FloatTensor): ``(batch_size, tgt_len, model_dim)``
            **kwargs
                enc_out: (only when using encoder/decoder)
                src_pad_mask: (only when using encoder/decoder)
                tgt_pad_mask (BoolTensor): ``(batch_size, tgt_len)``
                    used to build the attention mask
                step: int
                with_aligh: bool
                return_attention: bool, set by `attn_debug`
                position_embeddings (FloatTensor): rotary position encodings, if any

        Returns:
            (FloatTensor, FloatTensor):
            * emb: decoder_out ``(batch_size, tgt_len, model_dim)``
            * attns: standard and align if set
        """
        assert emb.dim() == 3  # batch x len x embedding_dim
        enc_out = kwargs.pop("enc_out", None)
        tgt_pad_mask = kwargs.pop("tgt_pad_mask", None)
        assert tgt_pad_mask is not None, "TransformerDecoder requires a tgt pad mask"
        src_pad_mask = kwargs.pop("src_pad_mask", None)
        if self.with_cross_attn:
            assert src_pad_mask is not None, "TransformerDecoder requires a src pad mask"
            src_pad_mask = src_pad_mask.unsqueeze(1)  # (batch x 1 x 1 x src_len)
            # dim 1 (heads) and 2 (tgt_len) will be broadcasted automatically in MHA
        left_pad = kwargs.pop("left_pad", False)
        step = kwargs.pop("step", None)
        with_align = kwargs.pop("with_align", False)
        return_attn = with_align or kwargs.pop("return_attn", False)
        position_embeddings = self.rope.update(emb.size(1), step=step)
        if self.rope_local is not None:
            position_embeddings_local = self.rope_local.update(emb.size(1), step=step)
        else:
            position_embeddings_local = position_embeddings
        decoder_in = kwargs.pop("decoder_in", None)
        image_token_id = kwargs.pop("image_token_id", None)
        prefix_len = kwargs.pop("prefix_len", None)
        attn_aligns = []

        if step == 0:
            self._enable_cache(emb.device, tgt_pad_mask)

        if emb.size(1) > 1:
            # training or first step decoding
            attn_mask = self._causal_attn_mask(tgt_pad_mask, prefix_len=prefix_len)
        else:
            # step by step decoding
            # we rebuild the pad mask of previous steps (left padded prompt)
            if left_pad:
                pad_mask = torch.zeros([emb.size(0), 1, step + 1], dtype=torch.bool, device=emb.device)
                pad_mask[:, :, : self.left_pad_mask.size(2)] = self.left_pad_mask
                attn_mask = ~pad_mask.unsqueeze(1)
            else:
                attn_mask = None

        # we need to adapt the mask for gemma3, TODO: find another condition?
        # SEEMS OK TO MASK IMAGES FOR LLAVA TOO ?
        if decoder_in is not None and attn_mask is not None:
            attn_mask = self._update_causal_mask(attn_mask, decoder_in == image_token_id)
        if self.sliding_window > 0 and (step or 0) >= self.sliding_window and attn_mask is not None:
            attn_mask = attn_mask[:, :, :, -self.sliding_window :]

        for i, layer in enumerate(self.transformer_layers):
            emb, attn = layer(
                emb,
                enc_out=enc_out if enc_out is not None else emb,
                src_pad_mask=src_pad_mask,
                attn_mask=attn_mask,
                step=step,
                return_attn=return_attn,
                position_embeddings=(
                    position_embeddings_local if (i + 1) % self.interleave_local else position_embeddings
                ),
            )
            if with_align:
                attn_align = layer.get_attn_align(
                    emb,
                    enc_out=enc_out if enc_out is not None else emb,
                    src_pad_mask=src_pad_mask,
                    attn_mask=~tgt_pad_mask,
                    step=step,
                    return_attn=return_attn,
                    position_embeddings=(
                        position_embeddings_local if (i + 1) % self.interleave_local else position_embeddings
                    ),
                    attns=attn,
                )
                if attn_align is not None:
                    attn_aligns.append(attn_align)

        emb = self.layer_norm(emb)

        # we take the first head
        top_attn = None if attn is None else attn[:, 0, :, :].contiguous()
        attns = {"std": top_attn}
        if with_align:
            attns["align"] = attn_aligns[self.alignment_layer]  # `(B, Q, K)`

        return emb, attns

    def _enable_cache(self, device, pad_mask):
        self.left_pad_mask = pad_mask
        for layer in self.transformer_layers:
            layer.self_attn.kcache = torch.empty(0, device=device)
            layer.self_attn.vcache = torch.empty(0, device=device)
            if layer.context_attn:
                layer.context_attn.kcache = torch.empty(0, device=device)
                layer.context_attn.vcache = torch.empty(0, device=device)

    def _disable_cache(self):
        self.left_pad_mask = None
        for layer in self.transformer_layers:
            layer.self_attn.kcache, layer.self_attn.vcache = None, None
            if layer.context_attn:
                layer.context_attn.kcache, layer.self_attn.vcache = None, None



"""Module for dynamic data transfrom."""

import os
import importlib

from .transform import (
    make_transforms,
    get_specials,
    TransformPipe,
    Transform,
)


AVAILABLE_TRANSFORMS = {}


def get_transforms_cls(transform_names):
    """Return valid transform class indicated in `transform_names`."""
    transforms_cls = {}
    for name in transform_names:
        if name not in AVAILABLE_TRANSFORMS:
            raise ValueError("%s transform not supported!" % name)
        transforms_cls[name] = AVAILABLE_TRANSFORMS[name]
    return transforms_cls


__all__ = [
    "get_transforms_cls",
    "get_specials",
    "make_transforms",
    "TransformPipe",
    "prepare_transforms",
]


def register_transform(name):
    """Transform register that can be used to add new transform class."""

    def register_transform_cls(cls):
        if name in AVAILABLE_TRANSFORMS:
            raise ValueError("Cannot register duplicate transform ({})".format(name))
        if not issubclass(cls, Transform):
            raise ValueError("transform ({}: {}) must extend Transform".format(name, cls.__name__))
        AVAILABLE_TRANSFORMS[name] = cls
        # populate name class attribute
        cls.name = name
        return cls

    return register_transform_cls


# Auto import python files in this directory
transform_dir = os.path.dirname(__file__)
for file in os.listdir(transform_dir):
    path = os.path.join(transform_dir, file)
    if not file.startswith("_") and not file.startswith(".") and (file.endswith(".py") or os.path.isdir(path)):
        file_name = file[: file.find(".py")] if file.endswith(".py") else file
        module = importlib.import_module("eole.transforms." + file_name)



#!/usr/bin/env python3
# coding: utf-8

import sacrebleu
import codecs
from eole.bin import BaseBin, register_bin


# Let's say you have a source file with N sentences in SL - eg: source.sl
# Translate your file in TL with the -n_best nbest options nbest being
# then number of hypotheses and output the target to -output target.nbest.tl
# This script can be run (for instance with nbest = 5) as follows:
# python mbr_bleu.py --nbest-hyp target.5.tl \
#       --nbest-order 5 --output target.mbr.tl
# It will compare all hyp with eachother and output the max bleu


def chunks(lgth, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(lgth), n):
        yield lgth[i : i + n]


@register_bin(name="mbr_bleu")
class MBRBleu(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument("--nbest-hyp", type=str, help="file with nbest to rerank", required=True)
        parser.add_argument("--nbest-order", type=int, help="nbest order", required=True)
        parser.add_argument("--output", type=str, help="output file", required=True)

    @classmethod
    def run(cls, args):
        with codecs.open(args.nbest_hyp, encoding="utf-8") as file:
            nbests = file.readlines()
            nbests = [item.strip() for item in nbests]
            nbests = chunks(nbests, args.nbest_order)

        with codecs.open(args.output, "w", encoding="utf-8") as output_file:
            best_indices = []
            for nbest in nbests:
                texts = []
                scores = []
                for hyp in nbest:
                    bleu = 0
                    for ref in nbest:
                        bleu = bleu + sacrebleu.sentence_bleu(hyp, [ref]).score / args.nbest_order
                        texts.append(hyp)
                        scores.append(bleu)
                max_index = scores.index(max(scores))
                output_file.write(texts[max_index] + "\n")
                best_indices.append(max_index)



import os
import yaml
import torch
import numpy as np
import codecs
from eole.utils.misc import set_random_seed
from eole.utils.logging import init_logger, logger
from eole.inputters.dynamic_iterator import build_dynamic_dataset_iter
from eole.utils.loss import LossCompute
from eole.constants import DefaultTokens, CorpusTask
from eole.transforms import get_transforms_cls

from argparse import ArgumentParser
from eole.bin import BaseBin, register_bin
from eole.config.cli import add_model
from eole.config import get_non_default_values
from eole.config.run import PredictConfig

"""
This script scores all sentences of a file using dynamic data.
For this purpose we use the same pipeline as the validation of a file
Below is an example of settings of a config.yaml file

model: lm-de.news2021_step_100000.pt
src: newstest2014-ref.de
tgt: newstest2014-ref.de
transforms: [onmt_tokenize]
batch_size: 16
gpu: 0
src_subword_type: bpe
src_subword_model: subwords.en_de.bpe
src_eoletok_kwargs: '{"mode": "aggressive"}'
tgt_subword_type: bpe
tgt_subword_model: subwords.en_de.bpe
tgt_eoletok_kwargs: '{"mode": "aggressive"}'

Output is the data and tab separated score
use the -output setting for preds + scores
Corpus PPL is in the logger.info
"""


@register_bin(name="LM_scoring")
class LMScoring(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "-config",
            "--config",
            "-c",
            required=False,
            help="Path of main YAML config file.",
        )

    @classmethod
    def run(cls, args):
        # logic copied from eole.bin.translate -> to refactor at some point
        if args.config is not None:
            with open(args.config) as f:
                config = yaml.safe_load(f)
        else:
            config = {}
        _parser = ArgumentParser()
        add_model(_parser, PredictConfig)
        defaults = vars(_parser.parse_args([]))
        stuff_to_update = get_non_default_values(args, defaults)
        config.update(stuff_to_update)
        config = PredictConfig(**config)
        init_logger(config.log_file)
        set_random_seed(config.seed, False)
        ppl_file = codecs.open(config.output + ".ppl", "w+", "utf-8")

        # no tensor_parallel support
        device = torch.device("cuda", config.gpu_ranks[0]) if len(config.gpu_ranks) > 0 else torch.device("cpu")
        if len(config.gpu_ranks) > 1:
            logger.warning(f"gpu_ranks is {str(config.gpu_ranks)} but only the first one will be used.")

        vocabs, model, model_opt = config.model.model_class.load_test_model(config)
        pad_token = vocabs["specials"].get("pad_token", DefaultTokens.PAD)
        padding_idx = vocabs["tgt"][pad_token]
        criterion = torch.nn.CrossEntropyLoss(ignore_index=padding_idx, reduction="none")
        valid_loss = LossCompute(
            criterion,
            model.generator,
            tgt_shift_index=0,
            lambda_coverage=model_opt.lambda_coverage,
            lambda_align=model_opt.lambda_align,
        )
        valid_loss.to(device)

        transforms_cls = get_transforms_cls(config._all_transform)
        infer_iter = build_dynamic_dataset_iter(
            args,
            transforms_cls,
            vocabs,
            task=CorpusTask.INFER,
            device_id=config.gpu,
        )

        model.to(device)
        model.eval()

        cumul_loss = 0.0
        cumul_length = 0
        # Now we can pipe the full file through the model using the Iterator

        for i, (batch, bucket_idx) in enumerate(infer_iter):
            # reminder a batch includes .src .tgt .indices and it is sorted
            batch_size = len(batch["srclen"])
            src = batch["src"]
            src_len = batch["srclen"]
            # print(batch)
            outputs, attns = model(src, None, src_len, with_align=False)
            # Compute and retrieve the loss for EACH sentence
            loss, _ = valid_loss(batch, outputs, attns)
            loss = loss.view(batch_size, -1)  # (B, T)
            losspertoken = loss.sum(1) / batch["tgt"][:, 1:, 0].ne(padding_idx).sum(1)
            ppl = torch.exp(losspertoken)
            cumul_loss += loss.sum().item()
            cumul_length += batch["tgt"][:, 1:, 0].ne(padding_idx).sum().cpu()
            # Now we need to rearrange the batch of ppl
            # in the original order with indices
            sent_ppl_orig = ppl.gather(
                0,
                torch.tensor(
                    sorted(
                        range(len(batch["cid_line_number"])),
                        key=lambda k: batch["cid_line_number"][k],
                    ),
                    device=ppl.device,
                ),
            )
            for j in range(batch_size):
                ppl_file.write(str(sent_ppl_orig[j].item()) + "\n")
        logger.info(
            "Loss: %.2f Tokens: %d Corpus PPL: %.2f" % (cumul_loss, cumul_length, np.exp(cumul_loss / cumul_length))
        )
        ppl_file.close()

        os.system('paste "' + config.src + '" "' + config.output + '".ppl > "' + config.output + '"')



"""
This includes: LossComputeBase and the standard NMTLossCompute, and
               sharded loss compute stuff.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import eole
from eole.modules.sparse_losses import SparsemaxLoss
from eole.modules.sparse_activations import LogSparsemax
from eole.constants import DefaultTokens
from eole.models.model import DecoderModel
from eole.utils.misc import get_device
from eole.utils.attention_entropy import compute_batch_attention_entropy

try:
    import ctranslate2
except ImportError:
    pass  # this is tested when importing for loading a LM


class LossCompute(nn.Module):
    """
    Class for managing efficient loss computation. Handles
    accumulating multiple loss computations.

    Args:
        criterion (:obj:`nn. loss function`) : NLLoss or customed loss
        generator (:obj:`nn.Module`) :
        lambda_coverage: Hyper-param to apply coverage attention if any
        lambda_align: Hyper-param for alignment loss
        tgt_shift_index (int): 1 for NMT, 0 for LM
        vocabs: full vocabs with specials
             module that maps the output of the decoder to a
             distribution over the target vocabulary.
        lm_generator (:obj:`ctranslate2.Generator`): LM Generator
        lm_prior_lambda (float): weight of LM model in loss
        lm_prior_tau (float): scaler for LM loss
    """

    def __init__(
        self,
        criterion,
        generator,
        lambda_coverage=0.0,
        lambda_align=0.0,
        tgt_shift_index=1,
        vocabs=None,
        lm_generator=None,
        lm_prior_lambda=None,
        lm_prior_tau=None,
        lm_prior_model=None,
    ):
        super(LossCompute, self).__init__()
        self.criterion = criterion
        self.generator = generator
        self.lambda_coverage = lambda_coverage
        self.lambda_align = lambda_align
        self.tgt_shift_index = tgt_shift_index
        self.vocabs = vocabs
        self.lm_generator = lm_generator
        self.lm_prior_lambda = lm_prior_lambda
        self.lm_prior_tau = lm_prior_tau
        self.lm_prior_model = lm_prior_model
        self.estimloss = nn.MSELoss(reduction="sum")

        self.pad_token = self.vocabs["specials"].get("pad_token", DefaultTokens.PAD)
        self.unk_token = self.vocabs["specials"].get("unk_token", DefaultTokens.UNK)
        self.eos_token = self.vocabs["specials"].get("eos_token", DefaultTokens.EOS)

    @classmethod
    def from_config(cls, config, model, vocabs, train=True):
        """
        Returns a subclass which wraps around an nn.Module subclass
        (such as nn.NLLLoss) which defines the loss criterion. The LossCompute
        object passes relevant data to a Statistics object which handles
        training/validation logging.
        The Criterion and LossCompute options are triggered by opt settings.
        """

        device = torch.device(get_device() if eole.utils.misc.use_gpu(config.training) else "cpu")
        pad_token = vocabs["specials"].get("pad_token", DefaultTokens.PAD)
        padding_idx = vocabs["tgt"][pad_token]

        if config.model.decoder is not None:
            lambda_align = getattr(
                config.model.decoder, "lambda_align", 0.0
            )  # patch to support non transformer configs
            if config.model.decoder.lambda_coverage != 0:
                lambda_coverage = config.model.decoder.lambda_coverage
                assert config.model.decoder.coverage_attn, (
                    "--coverage_attn needs to be set in " "order to use --lambda_coverage != 0"
                )
            else:
                lambda_coverage = 0
        else:
            lambda_coverage = 0
            lambda_align = 0.0

        tgt_shift_idx = model.tgt_shift

        if config.model.generator_function == "sparsemax":
            criterion = SparsemaxLoss(ignore_index=padding_idx, reduction="sum")
        else:
            criterion = nn.CrossEntropyLoss(
                ignore_index=padding_idx,
                reduction="sum",
                label_smoothing=config.training.label_smoothing,
            )

        lm_prior_lambda = config.training.lm_prior_lambda
        lm_prior_tau = config.training.lm_prior_tau
        if config.training.lm_prior_model:
            if config.training.lm_prior_model[-3:] == ".pt":
                _, lm_prior_model, lm_model_config = DecoderModel.load_test_model(
                    config, model_path=config.training.lm_prior_model
                )  # lm_model_config does not seem used
                # lm_prior_model.to(torch.device("cuda", config.training.gpu))
                lm_prior_model.eval()
                lm_generator = None
            else:
                lm_prior_model = None
                try:
                    import ctranslate2

                    lm_generator = ctranslate2.Generator(
                        config.training.lm_prior_model,
                        device="cuda",
                        compute_type="float16",
                    )
                except ImportError:
                    raise ImportError("Could not import ctranslate2")
        else:
            lm_generator = None
            lm_prior_model = None

        compute = cls(
            criterion,
            model.generator,
            lambda_coverage=lambda_coverage,
            lambda_align=lambda_align,
            tgt_shift_index=tgt_shift_idx,
            vocabs=vocabs,
            lm_generator=lm_generator,
            lm_prior_lambda=lm_prior_lambda,
            lm_prior_tau=lm_prior_tau,
            lm_prior_model=lm_prior_model,
        )

        # Store attention entropy configuration
        compute.log_attention_entropy = getattr(config.training, "log_attention_entropy", False)
        compute.attention_entropy_types = getattr(config.training, "attention_entropy_types", None)
        compute.attention_entropy_layers = getattr(config.training, "attention_entropy_layers", None)
        compute.attention_entropy_aggregation = getattr(config.training, "attention_entropy_aggregation", "mean")

        compute.to(device)  # this sometimes make embeddings move to the wrong device (cpu), not sure why

        return compute

    @property
    def padding_idx(self):
        return self.criterion.ignore_index

    def _compute_coverage_loss(self, std_attn, cov_attn, tgt):
        """compute coverage loss"""
        zero_attn = torch.zeros(cov_attn.size()[1:], device=cov_attn.device)
        cov_attn = torch.cat((zero_attn.unsqueeze(0), cov_attn[:-1]), 0)
        covloss = torch.min(std_attn, cov_attn).sum(dim=-1).view(-1)

        covloss[tgt == self.padding_idx] = 0
        return covloss.sum()

    def _compute_alignement_loss(self, align_head, ref_align):
        """Compute loss between 2 partial alignment matrix."""
        # align_head contains value in [0, 1) presenting attn prob,
        # 0 was resulted by the context attention src_pad_mask
        # So, the correspand position in ref_align should also be 0
        # Therefore, clip align_head to > 1e-18 should be bias free.
        align_loss = -align_head.clamp(min=1e-18).log().mul(ref_align).sum()
        align_loss *= self.lambda_align
        return align_loss

    def _compute_lm_loss_ct2(self, output, target):
        """
        Compute the loss between MT output and LM output
        https://github.com/cbaziotis/lm-prior-for-nmt/blob/master
        /fairseq_extension/user/lm_prior/lm_prior.py#L131-L133
        """

        # rescale with tau (temperature) and apply the log_softmax.
        scores = self.generator(self._bottle(output)) / self.lm_prior_tau
        scores = F.log_softmax(scores.to(torch.float32), dim=-1)

        src = target.detach().clone()
        src[src == self.vocabs["tgt"][self.eos_token]] = self.padding_idx
        src = src[:, :-1, :]
        src_len = src[:, :, 0].ne(self.padding_idx).sum(1)
        # ct2 expects src with lengths without padding
        lm_scores = self.lm_generator.forward_batch(
            ctranslate2.StorageView.from_array(src[:, :, 0].to(torch.int32)),
            ctranslate2.StorageView.from_array(src_len.to(torch.int32)),
            return_log_probs=False,
        )
        lm_scores = torch.as_tensor(lm_scores, device=scores.device)
        # again we use raw probs to rescale with tau and apply log_softmax
        lm_scores = self._bottle(lm_scores) / self.lm_prior_tau
        lm_scores = F.log_softmax(lm_scores.to(torch.float32), dim=-1)
        lm_scores[:, self.vocabs["tgt"]["unk_token"]] = -50
        lm_scores[:, self.vocabs["tgt"]["eos_token"]] -= 20
        # lm_scores are in log space so log_target=True
        lm_loss = F.kl_div(scores, lm_scores, reduction="none", log_target=True).sum(-1)
        non_padding = self._bottle(output).ne(self.padding_idx)[:, 0]
        lm_loss = lm_loss.masked_select(non_padding).sum()
        lm_loss = lm_loss * (self.lm_prior_tau**2)
        return lm_loss

    def _compute_lm_loss(self, output, target):
        """
        Compute the loss between MT output and LM output
        https://github.com/cbaziotis/lm-prior-for-nmt/blob/master
        /fairseq_extension/user/lm_prior/lm_prior.py#L131-L133
        """
        # rescale with tau (temperature) and apply the log_softmax.
        scores = self.generator(self._bottle(output)) / self.lm_prior_tau
        scores = F.log_softmax(scores.to(torch.float32), dim=-1)

        src = target.detach().clone()
        src[src == self.vocabs["tgt"]["eos_token"]] = self.padding_idx
        src = src[:, :-1, :]
        src_len = src[:, :, 0].ne(self.padding_idx).sum(1)
        # ct2 expects src with lengths without padding
        lm_outs, _ = self.lm_prior_model(src, None, src_len, with_align=False)
        lm_scores = self.lm_prior_model.generator(self._bottle(lm_outs)).detach().clone() / self.lm_prior_tau
        # again we use raw probs to rescale with tau and apply log_softmax
        lm_scores = F.log_softmax(lm_scores.to(torch.float32), dim=-1)
        lm_scores[:, self.vocabs["tgt"]["unk_token"]] = -50
        lm_scores[:, self.vocabs["tgt"]["eos_token"]] -= 20
        # lm_scores are in log space so log_target=True
        lm_loss = F.kl_div(scores, lm_scores, reduction="none", log_target=True).sum(-1)
        non_padding = self._bottle(output).ne(self.padding_idx)[:, 0]
        lm_loss = lm_loss.masked_select(non_padding).sum()
        lm_loss = lm_loss * (self.lm_prior_tau**2)
        return lm_loss

    def _bottle(self, _v):
        return _v.view(-1, _v.size(2))

    def _unbottle(self, _v, batch_size):
        return _v.view(-1, batch_size, _v.size(1))

    def ignore_prompt(self, batch):
        """
        Mask the prompt in the target side of the batch examples in order
            to set the loss of the prompt to zero.
        For finetuning on specific tasks.
        The end of the prompt must be indicated by `the DefaultTokens.MASK_BEFORE`
            placeholder.
        The masks are supposed to be properly handled by the loss criterion
            (e.g. nn.CrossEntropyLoss ).

        Args:
            batch: The current batch.
        """
        # Create a mask with zeros at prompt positions and ones at answer postions.
        mask = batch["src"].squeeze(dim=-1) == self.padding_idx
        mask = torch.cumsum(mask.int(), 1)
        # Apply the mask on the target side.
        batch["tgt"] *= mask.int()
        # Put the padding token index at the prompt positions.
        batch["tgt"] += self.padding_idx * (1 - mask.int())
        return batch

    def forward(self, batch, output, attns, estim=None):
        """Compute the forward loss

        Args:
          batch (batch) : batch of labeled examples
          output (:obj:`FloatTensor`) :
              output of decoder model ``(batch, tgt_len, hidden)``
          attns (dict) : dictionary of attention weights
              ``(batch, tgt_len, src_len)``

        Returns:
            A tuple with the loss and a :obj:`eole.utils.Statistics` instance.
        """
        # take into account here the tgt_shift_index (0 / 1 = LM/NMT)
        flat_tgt = batch["tgt"][:, self.tgt_shift_index :].contiguous().view(-1)

        if self.generator is not None:
            scores = self.generator(self._bottle(output))
            if isinstance(self.criterion, SparsemaxLoss):
                scores = LogSparsemax(scores.to(torch.float32), dim=-1)
            loss = self.criterion(scores.to(torch.float32), flat_tgt)
        else:
            loss = torch.tensor([0.0], device=output.device)
            scores = None

        if self.lambda_align != 0.0:
            align_head = attns["align"]
            if align_head.dtype != loss.dtype:  # Fix FP16
                align_head = align_head.to(loss.dtype)
            align_idx = batch["align"]
            batch_size, pad_tgt_size = batch["tgt"].size()
            _, pad_src_size = batch["src"].size()
            align_matrix_size = [batch_size, pad_tgt_size, pad_src_size]
            ref_align = eole.utils.make_batch_align_matrix(align_idx, align_matrix_size, normalize=True)
            ref_align = ref_align[:, self.tgt_shift_index :, :]
            if ref_align.dtype != loss.dtype:
                ref_align = ref_align.to(loss.dtype)
            align_loss = self._compute_alignement_loss(align_head=align_head, ref_align=ref_align)
            loss += align_loss

        if self.lambda_coverage != 0.0:
            coverage_loss = self._compute_coverage_loss(attns["std"], attns["coverage"], flat_tgt)
            loss += coverage_loss

        if self.lm_generator is not None:
            lm_loss = self._compute_lm_loss_ct2(output, batch["tgt"])
            loss = loss + lm_loss * self.lm_prior_lambda

        if self.lm_prior_model is not None:
            lm_loss = self._compute_lm_loss(output, batch["tgt"])
            loss = loss + lm_loss * self.lm_prior_lambda

        if estim is not None:
            batch["sco"] = batch["sco"].to(estim.dtype)
            estimloss = self.estimloss(estim, batch["sco"]).to(estim.dtype)
        else:
            estimloss = torch.tensor([0.0], device=loss.device)
        n_sents = len(batch["srclen"])

        # Compute attention entropy if attention weights are available and enabled
        attention_entropy = 0.0
        attention_available = attns and any(attn is not None for attn in attns.values())
        if hasattr(self, "log_attention_entropy") and self.log_attention_entropy and attention_available:
            try:
                attention_entropy = compute_batch_attention_entropy(
                    attns,
                    attention_types=getattr(self, "attention_entropy_types", None),
                    layer_indices=getattr(self, "attention_entropy_layers", None),
                    aggregation_method=getattr(self, "attention_entropy_aggregation", "mean"),
                )
            except Exception:
                # If entropy computation fails, default to 0
                attention_entropy = 0.0

        stats = self._stats(
            n_sents,
            loss.sum().item(),
            estimloss.item(),
            scores,
            flat_tgt,
            batch["cid"],
            batch["cid_line_number"],
            attention_entropy,
        )

        return loss, stats, estimloss

    def _stats(self, bsz, loss, auxloss, scores, target, cids, cids_idx, attention_entropy=0.0):
        """
        Args:
            loss (int): the loss computed by the loss criterion.
            scores (:obj:`FloatTensor`): a score for each possible output
            target (:obj:`FloatTensor`): true targets
            attention_entropy (float): computed attention entropy for this batch

        Returns:
            :obj:`eole.utils.Statistics` : statistics for this batch.
        """
        non_padding = target.ne(self.padding_idx)
        num_non_padding = non_padding.sum().item()
        if scores is not None:
            pred = scores.max(1)[1]
            num_correct = pred.eq(target).masked_select(non_padding).sum().item()
        else:
            num_correct = 0
        n_batchs = 1 if bsz else 0
        # in the case criterion reduction is None then we need
        # to sum the loss of each sentence in the batch
        data = {}
        for cid, idx in zip(cids, cids_idx):
            if cid not in data.keys():
                data[cid] = {"count": 1, "index": idx}
            else:
                data[cid]["count"] += 1
                data[cid]["index"] = min(idx, data[cid]["index"])
        return eole.utils.Statistics(
            loss=loss,
            auxloss=auxloss,
            n_batchs=n_batchs,
            n_sents=bsz,
            n_tokens=num_non_padding,
            n_correct=num_correct,
            data_stats=data,
            attention_entropy=attention_entropy * bsz,  # Scale by batch size
            n_attention_samples=bsz,
        )



import os
import torch
import json
import shutil
from collections import deque
from eole.utils.logging import logger
from eole.inputters.inputter import vocabs_to_dict
from eole.modules.lora import lora_state_dict
from eole.config import recursive_model_fields_set
from eole.config.run import TrainConfig
from eole.constants import DefaultTokens

try:
    from safetensors.torch import save_file
except ImportError:
    raise ImportError("run: pip install safetensors, to use safetensors")


def build_model_saver(config, model, vocabs, optim, device_id, transforms):

    model_saver = TrainingModelSaver(
        # config.training.save_model,
        config.training.model_path,
        model,
        config,
        vocabs,
        optim,
        config.training.keep_checkpoint,
        config.training.save_format,
        device_id,
        transforms,
    )
    return model_saver


def load_checkpoint(model_path):
    """
    Load checkpoint from `model_path` if any else return `None`.
    TODO: we need to clarify this compared to eole.models.BaseModel.load_checkpoint
    """
    checkpoint = None
    if not os.path.exists(model_path):
        raise FileNotFoundError(f"{model_path} does not seem to exist.")
    elif os.path.isdir(model_path):
        os.environ["MODEL_PATH"] = model_path
        logger.info("Loading checkpoint from %s" % model_path)
        # checkpoint = torch.load(ckpt_path, map_location=torch.device("cpu"))
        checkpoint = {}
        config_path = os.path.join(model_path, "config.json")
        if os.path.exists(config_path):
            with open(config_path) as f:
                config_dict = json.loads(os.path.expandvars(f.read()))
                # drop data to prevent validation issues
                config_dict["data"] = {}
                _config = TrainConfig(**config_dict)
                checkpoint["config"] = _config
        else:
            raise FileNotFoundError(f"{model_path} does not contain config.json")
        vocab_path = os.path.join(model_path, "vocab.json")
        if os.path.exists(vocab_path):
            with open(vocab_path, encoding="utf-8") as f:
                checkpoint["vocab"] = json.load(f)
            # use default specials if not specified
            if "specials" not in checkpoint["vocab"].keys():
                checkpoint["vocab"]["specials"] = {
                    "bos_token": DefaultTokens.BOS,
                    "pad_token": DefaultTokens.PAD,
                    "eos_token": DefaultTokens.EOS,
                    "unk_token": DefaultTokens.UNK,
                }
        else:
            raise FileNotFoundError(f"{model_path} does not contain vocab.json")
        optim_path = os.path.join(model_path, "optimizer.pt")
        if os.path.exists(optim_path):
            checkpoint["optim"] = torch.load(optim_path, map_location=torch.device("cpu"), weights_only=True)
    else:
        raise FileNotFoundError(f"{model_path} is not a directory.")

    return checkpoint


class ModelSaverBase(object):
    """
    Base class for various model saving scenarii.
    Not really used for now, but will be when ConvertModelSaver is properly implemented.
    """

    def __init__(self, model_path, model, config, vocabs, device_id=0, transforms=None):
        self.model_path = model_path
        self.model = model
        self.config = config
        self.vocabs = vocabs
        self.device_id = device_id
        self.transforms = transforms

    def _save(self, *args, **kwargs):
        raise NotImplementedError()


# TODO: move saving and sharding logic from convert tools here
# class ConvertModelSaver(ModelSaverBase):
#     pass


class TrainingModelSaver(ModelSaverBase):
    """Simple model saver to filesystem"""

    def __init__(
        self,
        model_path,
        model,
        config,
        vocabs,
        optim=None,
        keep_checkpoint=-1,
        save_format="pytorch",
        device_id=0,
        transforms=None,
    ):
        self.model_path = model_path
        self.step_dir = None
        self.model = model
        self.config = config
        self.vocabs = vocabs
        self.optim = optim
        self.last_saved_step = None
        self.keep_checkpoint = keep_checkpoint
        self.save_format = save_format
        self.device_id = device_id
        self.transforms = transforms

        # we might raise some warning if model_path already exists
        if not os.path.exists(self.model_path):
            os.makedirs(self.model_path, exist_ok=True)

        if keep_checkpoint > 0:
            self.checkpoint_queue = deque([], maxlen=keep_checkpoint)

    def cleanup(self):
        if self.keep_checkpoint > 0:
            if len(self.checkpoint_queue) == self.checkpoint_queue.maxlen:
                step_dir_to_delete = os.path.join(self.model_path, self.checkpoint_queue.popleft())
                try:
                    shutil.rmtree(step_dir_to_delete)
                except FileNotFoundError:
                    pass
                except Exception:
                    raise

            self.checkpoint_queue.append(self.step_dir)

    def _maybe_lora(self, model):
        if (hasattr(self.config.training, "lora_layers") and len(self.config.training.lora_layers) > 0) or (
            hasattr(self.config.training, "lora_embedding") and self.config.training.lora_embedding
        ):
            model_state_dict = lora_state_dict(model, bias="lora_only")
            for k, v in model.state_dict().items():
                if "estimator" in k:
                    model_state_dict[k] = v
        else:
            model_state_dict = model.state_dict()
        return model_state_dict

    def _tensor_parallel_state_dict(self, model_state_dict, world_size):
        full_model = [None for _ in range(world_size)]
        for key, value in model_state_dict.items():
            model_state_dict[key] = value.cpu()
        torch.distributed.all_gather_object(full_model, model_state_dict)
        full_state_dict = {}
        for key in full_model[0].keys():
            key_2, key_1 = key.split(".")[-2:]
            averaged_params = {
                "linear_keys",
                "linear_values",
                "linear_query",
                "gate_up_proj",
                "up_proj",
            }
            cat_params = {"final_linear", "down_proj"}
            # we probably should try and improve this to rely on dimensions instead of names
            match key_1, key_2:
                case "lora_A", _ if key_2 in averaged_params:
                    full_state_dict[key] = sum([full_model[i][key].cpu() for i in range(world_size)]) / world_size
                case "lora_A", _ if key_2 in cat_params:
                    full_state_dict[key] = torch.cat([full_model[i][key].cpu() for i in range(world_size)], 1)
                case "lora_B", _ if key_2 in averaged_params:
                    full_state_dict[key] = torch.cat([full_model[i][key].cpu() for i in range(world_size)], 0)
                case "lora_B", _ if key_2 in cat_params:
                    full_state_dict[key] = torch.cat([full_model[i][key].cpu() for i in range(world_size)], 1)
                case _ if key_1 in averaged_params:
                    full_state_dict[key] = torch.cat([full_model[i][key].cpu() for i in range(world_size)], 0)
                case _ if key_1 in cat_params:
                    full_state_dict[key] = torch.cat([full_model[i][key].cpu() for i in range(world_size)], 1)
                case _ if key_2 in averaged_params:
                    full_state_dict[key] = torch.cat([full_model[i][key].cpu() for i in range(world_size)], 0)
                case _ if key_2 in cat_params:
                    full_state_dict[key] = torch.cat([full_model[i][key].cpu() for i in range(world_size)], 1)
                case _, _:
                    full_state_dict[key] = full_model[0][key]
        return full_state_dict

    def _save_tokenizer(self):
        pass

    def update_step_dir(self, step):
        self.step_dir = f"step_{step}"
        step_dir = os.path.join(self.model_path, self.step_dir)
        if not os.path.exists(step_dir):
            os.mkdir(step_dir)

    def _make_symlink(self, item):
        # symlink item from step_X dir to parent model_path
        symlink_src_path = os.path.join(self.step_dir, item)
        symlink_dst_path = os.path.join(self.model_path, item)
        if os.path.islink(symlink_dst_path):
            os.remove(symlink_dst_path)
        elif os.path.exists(symlink_dst_path):
            os.remove(symlink_dst_path)
        try:
            os.symlink(symlink_src_path, symlink_dst_path)
        except PermissionError:
            # If a symlink cannot be created, copy the file instead
            shutil.copy2(os.path.join(self.model_path, symlink_src_path), symlink_dst_path)

    def _save_optimizer(self):
        optim_data = self.optim.state_dict()
        optim_path = os.path.join(self.model_path, self.step_dir, "optimizer.pt")
        torch.save(optim_data, optim_path)
        self._make_symlink("optimizer.pt")

    def _save_weights(self, model_state_dict):
        # we might implement sharding later
        model_path = os.path.join(self.model_path, self.step_dir, "model.00.safetensors")
        if self.config.model.share_embeddings and "tgt_emb.embeddings.weight" in model_state_dict:
            model_state_dict.pop("tgt_emb.embeddings.weight")
        if self.config.model.share_decoder_embeddings and "generator.weight" in model_state_dict:
            model_state_dict.pop("generator.weight")
        if self.config.training.freeze_encoder:
            model_state_dict = {
                k: v for k, v in model_state_dict.items() if not (k.startswith("encoder") or k.startswith("adapter"))
            }
        if self.config.training.freeze_decoder:
            model_state_dict = {
                k: v
                for k, v in model_state_dict.items()
                if not (k.startswith("decoder") or k.startswith("generator") or k.startswith("tgt"))
            }
        save_file(model_state_dict, model_path)
        self._make_symlink("model.00.safetensors")

    def _save_vocab(self):
        vocab_data = vocabs_to_dict(self.vocabs)
        vocab_path = os.path.join(self.model_path, self.step_dir, "vocab.json")
        with open(vocab_path, "w", encoding="utf-8") as f:
            json.dump(vocab_data, f, indent=2, ensure_ascii=False)
        self._make_symlink("vocab.json")

    def _save_config(self):
        config_data = recursive_model_fields_set(self.config)
        config_path = os.path.join(self.model_path, self.step_dir, "config.json")
        with open(config_path, "w", encoding="utf-8") as f:
            json.dump(config_data, f, indent=2, ensure_ascii=False)
        self._make_symlink("config.json")

    def _save_transforms_artifacts(self):
        if self.transforms is not None:
            checkpoint_path = os.path.join(self.model_path, self.step_dir)
            for transform_name, transform in self.transforms.items():
                transform_save_config, artifacts = transform._save_artifacts(checkpoint_path)
                setattr(
                    self.config.transforms_configs,
                    transform_name,
                    transform_save_config,
                )
                for artifact in artifacts:
                    self._make_symlink(artifact)
                # we probably do not need to save transforms artifacts for each checkpoint
                # transform._save_artifacts(os.path.join(self.model_path, self.step_dir))

    # previously _st_save
    def _save(self, step):
        # convert state_dict to lora if needed
        model_state_dict = self._maybe_lora(self.model)

        if torch.distributed.is_initialized():
            world_size = torch.distributed.get_world_size()
        else:
            world_size = 1

        if world_size > 1:
            # gather model split across GPUs for tensor parallelism
            model_state_dict = self._tensor_parallel_state_dict(model_state_dict, world_size)

        if not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0:
            self.update_step_dir(step)
            logger.info(f"Saving optimizer and weights to {self.step_dir}, and symlink to {self.model_path}")
            self._save_optimizer()
            self._save_weights(model_state_dict)
            logger.info("Saving transforms artifacts, if any, " f"to {os.path.join(self.model_path, self.step_dir)}")
            self._save_transforms_artifacts()
            logger.info(f"Saving config and vocab to {self.model_path}")
            self._save_vocab()
            self._save_config()
            self.cleanup()
            # we shall trigger optional saves from transforms here + some default inference config ?
        if torch.distributed.is_initialized():
            torch.distributed.barrier()

    def save(self, step, moving_average=None):
        """Main entry point for model saver

        It wraps the `_save` method with checks and apply `keep_checkpoint`
        related logic
        """

        if self.keep_checkpoint == 0 or step == self.last_saved_step:
            return

        if moving_average:
            model_params_data = []
            for avg, param in zip(moving_average, self.model.parameters()):
                model_params_data.append(param.data)
                param.data = avg.data

        self._save(step)

        self.last_saved_step = step

        if moving_average:
            for param_data, param in zip(model_params_data, self.model.parameters()):
                param.data = param_data



"""Define a minimal encoder."""

from eole.encoders.encoder import EncoderBase
import torch


class MeanEncoder(EncoderBase):
    """A trivial non-recurrent encoder. Simply applies mean pooling.

    Args:
        encoder_config (eole.config.ModelConfig)
        embeddings (eole.modules.Embeddings):
          embeddings to use, should have positional encodings
        running_config (TrainingConfig / InferenceConfig)
    """

    def __init__(self, encoder_config, running_config=None):
        super(MeanEncoder, self).__init__()
        self.num_layers = encoder_config.layers

    @classmethod
    def from_config(cls, encoder_config, running_config=None):
        """Alternate constructor."""
        # config = opt.model.encoder  # MeanEncoderConfig
        return cls(encoder_config)

    def forward(self, emb, **kwargs):
        """See :func:`EncoderBase.forward()`"""
        pad_mask = kwargs.pop("pad_mask", None)
        batch, _, emb_dim = emb.size()

        if pad_mask is not None:
            # we avoid padding while mean pooling
            pad_mask = (~pad_mask).float()
            mean = torch.bmm(pad_mask, emb).mean(1)
        else:
            mean = emb.mean(1)

        mean = mean.expand(self.num_layers, batch, emb_dim)
        enc_out = emb
        enc_final_hs = (mean, mean)
        return enc_out, enc_final_hs



#!/usr/bin/env python
"""Train models with dynamic data."""
import torch
from functools import partial
from eole.utils.distributed import ErrorHandler
from eole.utils.distributed_workers import spawned_train
from eole.utils.misc import set_random_seed
from eole.utils.logging import init_logger, logger
from argparse import ArgumentParser

from eole.train_single import main as single_main

from eole.config.run import TrainConfig
from eole.bin import register_bin
from eole.bin.run import RunBin


# Set sharing strategy manually instead of default based on the OS.
# torch.multiprocessing.set_sharing_strategy('file_system')


def train(config):
    # config is the full TrainConfig namespace here
    config._validate_vocab_config()
    init_logger(config.log_file)
    set_random_seed(config.seed, False)

    train_process = partial(single_main)

    # a running config can either be training or inference, here we are training
    running_config = config.training

    nb_gpu = len(running_config.gpu_ranks)

    if running_config.world_size > 1:
        mp = torch.multiprocessing.get_context("spawn")
        # Create a thread to listen for errors in the child processes.
        error_queue = mp.SimpleQueue()
        error_handler = ErrorHandler(error_queue)
        # Train with multiprocessing.
        procs = []
        for device_id in range(nb_gpu):
            procs.append(
                mp.Process(
                    target=spawned_train,
                    args=(train_process, config, device_id, error_queue),
                    daemon=False,
                )
            )
            procs[device_id].start()
            logger.info(" Starting process pid: %d  " % procs[device_id].pid)
            error_handler.add_child(procs[device_id].pid)
        for p in procs:
            p.join()

    elif nb_gpu == 1:  # case 1 GPU only
        train_process(config, device_id=0)
    else:  # case only CPU
        train_process(config, device_id=-1)


@register_bin(name="train")
class Train(RunBin):
    config_class = TrainConfig
    require_config = True

    @classmethod
    def run(cls, args):
        config = cls.build_config(args)
        train(config)


def _get_parser():
    # Patch to allow autodoc with sphinx-argparse
    parser = ArgumentParser()
    Train.add_args(parser)
    return parser



"""Module defining various utilities."""

from eole.utils.misc import use_gpu, set_random_seed
from eole.utils.alignment import make_batch_align_matrix
from eole.utils.report_manager import ReportMgr, build_report_manager
from eole.utils.statistics import Statistics
from eole.utils.optimizers import MultipleOptimizer, Optimizer, AdaFactor
from eole.utils.attention_entropy import (
    compute_attention_entropy,
    compute_attention_entropy_from_dict,
    aggregate_attention_entropy,
    compute_batch_attention_entropy,
)
from eole.utils.earlystopping import EarlyStopping, scorers_from_config

__all__ = [
    "use_gpu",
    "set_random_seed",
    "ReportMgr",
    "build_report_manager",
    "Statistics",
    "MultipleOptimizer",
    "Optimizer",
    "AdaFactor",
    "EarlyStopping",
    "scorers_from_config",
    "make_batch_align_matrix",
    "compute_attention_entropy",
    "compute_attention_entropy_from_dict",
    "aggregate_attention_entropy",
    "compute_batch_attention_entropy",
]



"""
    Inspired from https://github.com/karpathy/llm.c/blob/master/dev/data/hellaswag.py
"""

import json
import yaml
import os
import requests
from tqdm import tqdm
from argparse import ArgumentParser
from eole.inference_engine import InferenceEnginePY

from eole.bin import BaseBin, register_bin
from eole.config.cli import add_model
from eole.config import get_non_default_values
from eole.config.run import PredictConfig

DATA_CACHE_DIR = os.path.join(os.path.dirname(__file__), "hellaswag")

hellaswags = {
    "train": "https://raw.githubusercontent.com/rowanz/hellaswag/master/data/hellaswag_train.jsonl",
    "val": "https://raw.githubusercontent.com/rowanz/hellaswag/master/data/hellaswag_val.jsonl",
    "test": "https://raw.githubusercontent.com/rowanz/hellaswag/master/data/hellaswag_test.jsonl",
}


def download_file(url: str, fname: str, chunk_size=1024):
    """Helper function to download a file from a given url"""
    resp = requests.get(url, stream=True)
    total = int(resp.headers.get("content-length", 0))
    with open(fname, "wb") as file, tqdm(
        desc=fname,
        total=total,
        unit="iB",
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for data in resp.iter_content(chunk_size=chunk_size):
            size = file.write(data)
            bar.update(size)


def download(split):
    """Downloads HellaSwag DATA_CACHE_DIR"""
    os.makedirs(DATA_CACHE_DIR, exist_ok=True)
    data_url = hellaswags[split]
    data_filename = os.path.join(DATA_CACHE_DIR, f"hellaswag_{split}.jsonl")
    if not os.path.exists(data_filename):
        print(f"Downloading {data_url} to {data_filename}...")
        download_file(data_url, data_filename)
    else:
        print(f"{data_filename} already exists, skipping download...")


def iterate_examples(split):
    # there are 10,042 examples in total in val
    download(split)
    with open(os.path.join(DATA_CACHE_DIR, f"hellaswag_{split}.jsonl"), "r") as f:
        for line in f:
            example = json.loads(line)
            yield example


def evaluate(config):
    total = 0
    correct = 0
    correct_norm = 0
    engine = InferenceEnginePY(config)
    left_pad = config.model.left_pad
    for data in iterate_examples("val"):
        total += 1
        # retrieve length of tokenized context
        transformed_context = engine.transform_pipe.apply({"src": data["ctx"].split(" "), "tgt": None})
        context_length = len(transformed_context["src"])
        label = data["label"]
        examples = [f'{data["ctx"]} {ending}' for ending in data["endings"]]
        scores = engine.score_list(examples)
        # do not consider context log probs (note: padding ones are 0)
        norm_scores = []
        max_len = max([len(log_probs) for _, log_probs, _ in scores])
        for _, log_probs, length in scores:
            if left_pad:
                offset = max_len - length + context_length
            else:
                offset = context_length
            norm_score = sum(log_probs[offset:]) / (length - context_length)
            norm_scores.append(norm_score)
        _scores = [item[0] for item in scores]
        pred_label = _scores.index(max(_scores))
        pred_label_norm = norm_scores.index(max(norm_scores))
        correct += int(pred_label == label)
        correct_norm += int(pred_label_norm == label)
        print(f"{total} acc: {correct/total:.4f} " f"acc_norm: {correct_norm}/{total}={correct_norm/total:.4f}")


@register_bin(name="eval_hellaswag")
class EvalHellaswag(BaseBin):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "-config",
            "--config",
            "-c",
            required=False,
            help="Path of main YAML config file.",
        )
        add_model(parser, PredictConfig)

    def run(args):
        # same as in eole.bin.translate, should be improved/factorized
        if args.config is not None:
            with open(args.config) as f:
                config = yaml.safe_load(os.path.expandvars(f.read()))
        else:
            config = {}
        _parser = ArgumentParser()
        add_model(_parser, PredictConfig)
        defaults = vars(_parser.parse_args([]))
        stuff_to_update = get_non_default_values(args, defaults)
        config.update(stuff_to_update)

        # pop extra fields added by argparse, not supported in pydantic configs
        if "bin" in config.keys():
            config.pop("bin")
        if "sub_bin" in config.keys():
            config.pop("sub_bin")
        if "config" in config.keys():
            config.pop("config")

        config = PredictConfig(**config)

        evaluate(config)
